# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#

from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html
from django.utils.http import urlencode

from hyperkitty import models


admin.site.register(models.profile.Profile)
admin.site.register(models.tag.Tag)
admin.site.register(models.vote.Vote)
admin.site.register(models.thread.LastView)
admin.site.register(models.favorite.Favorite)


@admin.register(models.email.Email)
class EmailAdmin(admin.ModelAdmin):
    list_display = ('id', 'sender', 'date', 'subject', 'mailinglist')
    list_filter = ('sender', 'mailinglist__list_id', 'thread_id')
    search_fields = ('subject', 'sender__address')


@admin.register(models.thread.Thread)
class ThreadAdmin(admin.ModelAdmin):
    list_display = ('id', 'mailinglist', 'subject', 'date_active', 'emails')
    search_fields = ('subject', 'mailinglist__name', 'mailinglist__list_id')
    list_filter = ('mailinglist__list_id', )

    def emails(self, obj):
        count = obj.emails.count()
        url = (
            reverse("admin:hyperkitty_email_changelist")
            + "?"
            + urlencode({"thread__id": f"{obj.id}"})
        )
        return format_html('<a href="{}">{} Emails</a>', url, count)


@admin.register(models.mailinglist.MailingList)
class MailingListAdmin(admin.ModelAdmin):
    list_display = ('name', 'list_id', 'threads', 'emails')
    search_fields = ('name', 'list_id')

    def threads(self, obj):
        count = obj.threads.count()
        url = (
            reverse("admin:hyperkitty_thread_changelist")
            + "?"
            + urlencode({"mailinglist__id": f"{obj.id}"})
        )
        return format_html('<a href="{}">{} Threads</a>', url, count)

    def emails(self, obj):
        count = obj.emails.count()
        url = (
            reverse("admin:hyperkitty_email_changelist")
            + "?"
            + urlencode({"mailinglist__id": f"{obj.id}"})
        )
        return format_html('<a href="{}">{} Emails</a>', url, count)


@admin.register(models.category.ThreadCategory)
class ThreadCategoryAdmin(admin.ModelAdmin):
    def save_model(self, request, obj, form, change):
        obj.name = obj.name.lower()
        return super(ThreadCategoryAdmin, self).save_model(
                     request, obj, form, change)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import logging
import os
import re
from email.generator import BytesGenerator
from email.message import EmailMessage
from email.utils import formataddr
from io import BytesIO

from django.conf import settings
from django.db import IntegrityError, models
from django.utils.timezone import get_fixed_timezone, now

from hyperkitty.lib.analysis import compute_thread_order_and_depth
from .common import VotesCachedValue
from .mailinglist import MailingList
from .thread import Thread
from .vote import Vote


logger = logging.getLogger(__name__)


class Email(models.Model):
    """
    An archived email, from a mailing-list. It is identified by both the list
    name and the message id.
    """
    mailinglist = models.ForeignKey(
        "MailingList", related_name="emails", on_delete=models.CASCADE)
    message_id = models.CharField(max_length=255, db_index=True)
    message_id_hash = models.CharField(max_length=255, db_index=True)
    sender = models.ForeignKey(
        "Sender", related_name="emails", on_delete=models.CASCADE)
    sender_name = models.CharField(max_length=255, null=True, blank=True)
    subject = models.CharField(max_length=512, db_index=True)
    content = models.TextField()
    date = models.DateTimeField(db_index=True)
    timezone = models.SmallIntegerField()
    in_reply_to = models.CharField(
        max_length=255, null=True, blank=True, db_index=True)
    # XXX(maxking): Delete behavior is handled by on_pre_delete(). Since we
    # manually make sure that we set and unset parents and handle the fact that
    # they can't be None for more than one Email in a thread, we don't want
    # database engine to enforce the relationship is valid. Hence, we set the
    # on_delete=DO_NOTHING and db_constraint=False.
    parent = models.ForeignKey(
        "self", blank=True, null=True, on_delete=models.DO_NOTHING,
        related_name="children", db_constraint=False)
    thread = models.ForeignKey(
        "Thread", related_name="emails", on_delete=models.CASCADE)
    archived_date = models.DateTimeField(default=now, db_index=True)
    thread_depth = models.IntegerField(default=0)
    thread_order = models.IntegerField(null=True, blank=True, db_index=True)

    ADDRESS_REPLACE_RE = re.compile(r"([\w.+-]+)@([\w.+-]+)")

    def __init__(self, *args, **kwargs):
        super(Email, self).__init__(*args, **kwargs)
        self.cached_values = {
            "votes": VotesCachedValue(self),
        }

    def __lt__(self, other):
        return self.date < other.date

    class Meta:
        unique_together = ("mailinglist", "message_id")

    def get_votes(self):
        return self.cached_values["votes"]()

    def vote(self, value, user):
        # Checks if the user has already voted for this message.
        existing = self.votes.filter(user=user).first()
        if existing is not None and existing.value == value:
            return  # Vote already recorded (should I raise an exception?)
        if value not in (0, 1, -1):
            raise ValueError("A vote can only be +1 or -1 (or 0 to cancel)")
        if existing is not None:
            # vote changed or cancelled
            if value == 0:
                existing.delete()
            else:
                existing.value = value
                existing.save()
        else:
            # new vote
            vote = Vote(email=self, user=user, value=value)
            vote.save()

    def set_parent(self, parent):
        if self.id == parent.id:
            raise ValueError("An email can't be its own parent")
        # Compute the subthread
        subthread = [self]

        def _collect_children(current_email):
            children = list(current_email.children.all())
            if not children:
                return
            subthread.extend(children)
            for child in children:
                _collect_children(child)
        _collect_children(self)
        # now set my new parent value
        old_parent_id = self.parent_id
        self.parent = parent
        self.save(update_fields=["parent_id"])
        # If my future parent is in my current subthread, I need to set its
        # parent to my current parent
        if parent in subthread:
            parent.parent_id = old_parent_id
            parent.save(update_fields=["parent_id"])
            # do it after setting the new parent_id to avoid having two
            # parent_ids set to None at the same time (IntegrityError)
        if self.thread_id != parent.thread_id:
            # we changed the thread, reattach the subthread
            former_thread = self.thread
            for child in subthread:
                child.thread = parent.thread
                child.save(update_fields=["thread_id"])
                if child.date > parent.thread.date_active:
                    parent.thread.date_active = child.date
            parent.thread.save()
            # if we were the starting email, or former thread may be empty
            if former_thread.emails.count() == 0:
                former_thread.delete()
        compute_thread_order_and_depth(parent.thread)

    def as_message(self, escape_addresses=True):
        # http://wordeology.com/computer/how-to-send-good-unicode-email-with-python.html
        # http://stackoverflow.com/questions/31714221/how-to-send-an-email-with-quoted
        # http://stackoverflow.com/questions/9403265/how-do-i-use-python/9509718#9509718
        msg = EmailMessage()

        # Headers
        def unfold(hdr):
            return re.sub('[\r\n]', '', hdr)
        unixfrom = "From %s %s" % (
            self.sender.address, self.archived_date.strftime("%c"))
        assert isinstance(self.sender.address, str)
        header_from = self.sender.address
        if self.sender_name and self.sender_name != self.sender.address:
            header_from = formataddr((self.sender_name, header_from))
        header_to = self.mailinglist.name
        msg.set_unixfrom(unixfrom)
        headers = (
            ("From", header_from),
            ("To", header_to),
            ("Subject", self.subject),
            )
        for header_name, header_value in headers:
            msg[header_name] = unfold(header_value)
        tz = get_fixed_timezone(self.timezone)
        header_date = self.date.astimezone(tz).replace(microsecond=0)
        # Date format: http://tools.ietf.org/html/rfc5322#section-3.3
        msg["Date"] = header_date.strftime("%a, %d %b %Y %H:%M:%S %z")
        # Clean bogus characters from Message-ID.
        msg["Message-ID"] = "<%s>" % re.sub(r'[\s<>()]', '', self.message_id)
        if self.in_reply_to:
            msg["In-Reply-To"] = "<%s>" % unfold(self.in_reply_to)

        # Body
        content = self.ADDRESS_REPLACE_RE.sub(r"\1(a)\2", self.content)

        # Enforce `multipart/mixed` even when there are no attachments.
        msg.set_content(content, subtype='plain')
        msg.make_mixed()

        # Attachments
        for attachment in self.attachments.order_by("counter"):
            mimetype = attachment.content_type.split('/', 1)
            msg.add_attachment(attachment.get_content(), maintype=mimetype[0],
                               subtype=mimetype[1], filename=attachment.name)

        return msg

    def as_bytes(self):
        """Similar to EmailMessage.as_bytes() but sets mangle_from_=True.

        Serialize the current email into bytes to be added into an mbox.
        """
        msg = self.as_message()
        # We have to call BytesGenerator here to set
        # mangle_from_=True, which is set to 'False' by default in
        # EmailMessage.as_bytes() method.
        fd = BytesIO()
        gen = BytesGenerator(fd, mangle_from_=True)
        gen.flatten(msg, unixfrom=True)
        return fd.getvalue()

    @property
    def display_fixed(self):
        return "@@" in self.content

    def _set_message_id_hash(self):
        from hyperkitty.lib.utils import get_message_id_hash  # circular import
        if not self.message_id_hash:
            self.message_id_hash = get_message_id_hash(self.message_id)

    def on_post_init(self):
        self._set_message_id_hash()

    def on_post_created(self):
        self.thread.on_email_added(self)
        self.mailinglist.on_email_added(self)
        if not getattr(settings, "HYPERKITTY_BATCH_MODE", False):
            # For batch imports, let the cron job do the work
            from hyperkitty.tasks import check_orphans
            check_orphans(self.id)

    def on_pre_save(self):
        self._set_message_id_hash()
        # Link to the thread
        if self.thread_id is None:
            # Create the thread if not found
            thread, _thread_created = Thread.objects.get_or_create(
                mailinglist=self.mailinglist,
                thread_id=self.message_id_hash)
            self.thread = thread
        # Make sure there is only one email with parent_id == None in a thread
        if self.parent_id is not None:
            return
        starters = Email.objects.filter(
                thread=self.thread, parent_id__isnull=True
            ).values_list("id", flat=True)
        if len(starters) > 0 and list(starters) != [self.id]:
            raise IntegrityError("There can be only one email with "
                                 "parent_id==None in the same thread")

    def on_post_save(self):
        pass

    def on_pre_delete(self):
        # Reset parent_id
        children = self.children.order_by("date")
        if not children:
            return
        if self.parent is None:
            #  Temporarily set the email's parent_id to not None, to allow the
            #  next email to be the starting email (there's a check on_save for
            #  duplicate thread starters)
            self.parent = self
            self.save(update_fields=["parent"])
            starter = children[0]
            starter.parent = None
            starter.save(update_fields=["parent"])
            children.all().update(parent=starter)
        else:
            children.update(parent=self.parent)

    def on_post_delete(self):
        try:
            thread = Thread.objects.get(id=self.thread_id)
        except Thread.DoesNotExist:
            pass
        else:
            thread.on_email_deleted(self)
        try:
            mlist = MailingList.objects.get(pk=self.mailinglist_id)
        except MailingList.DoesNotExist:
            pass
        else:
            mlist.on_email_deleted(self)

    def on_vote_added(self, vote):
        from hyperkitty.tasks import rebuild_email_cache_votes
        rebuild_email_cache_votes(self.id)

    on_vote_deleted = on_vote_added


class Attachment(models.Model):
    email = models.ForeignKey(
        "Email", related_name="attachments", on_delete=models.CASCADE)
    counter = models.SmallIntegerField()
    name = models.CharField(max_length=255)
    content_type = models.CharField(max_length=255)
    encoding = models.CharField(max_length=255, null=True)
    size = models.IntegerField(null=True)
    content = models.BinaryField(null=True)

    class Meta:
        unique_together = ("email", "counter")

    def on_pre_save(self):
        # set the size
        if not self.size and self.content is not None:
            self.size = len(self.content)

    def _get_folder(self):
        global_folder = getattr(
            settings, "HYPERKITTY_ATTACHMENT_FOLDER", None)
        if global_folder is None:
            return None
        mlist = self.email.mailinglist.name
        try:
            listname, domain = mlist.rsplit("@", 1)
        except ValueError:
            listname = "none"
            domain = mlist
        return os.path.join(
            global_folder, domain, listname,
            self.email.message_id_hash[0:2],
            self.email.message_id_hash[2:4],
            self.email.message_id_hash[4:6],
            str(self.email.id),
        )

    def get_content(self):
        folder = self._get_folder()
        if self.content is not None:
            return bytes(self.content)
        if folder is None:
            return ""
        filepath = os.path.join(folder, str(self.counter))
        if not os.path.exists(filepath):
            logger.error("Could not find local attachment %s for email %s",
                         self.counter, self.email.id)
            return ""
        with open(filepath, "rb") as f:
            content = f.read()
        return content

    def set_content(self, content):
        if isinstance(content, str):
            if self.encoding is not None:
                content = content.encode(self.encoding, errors='replace')
            else:
                content = content.encode('utf-8')
        self.size = len(content)
        folder = self._get_folder()
        if folder is None:
            self.content = content
            return
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(os.path.join(folder, str(self.counter)), "wb") as f:
            f.write(content)
        self.content = None



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#
from django.apps import AppConfig


class HyperKittyConfig(AppConfig):
    name = 'hyperkitty'
    verbose_name = "HyperKitty"

    def ready(self):
        # flake8: noqa
        import hyperkitty.checks
        import hyperkitty.signals



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>

"""
Find the parents of orphan emails.

When an email has an in-reply-to header pointing to no known email, it
is set to start a new thread. This job runs later on and tries to find the
missing email in the database, in case it was added in the meantime.

This can happen if HK receives the reply before the original message (when a
mail server in the chain has an issue, or in case of greylisting for example).
"""

from django_extensions.management.jobs import BaseJob

from hyperkitty.models import Email


class Job(BaseJob):
    help = "Reattach orphan emails"
    when = "daily"

    def execute(self):
        orphan_starters = Email.objects.filter(
            parent_id__isnull=True, in_reply_to__isnull=False)
        for orphan_starter in orphan_starters:
            parent = Email.objects.filter(
                mailinglist=orphan_starter.mailinglist,
                message_id=orphan_starter.in_reply_to).first()
            if parent is None:
                continue
            if orphan_starter.id == parent.id:
                # an email with the in-reply-to header pointing to itself,
                # that's just bogus, ignore it.
                continue
            orphan_starter.set_parent(parent)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import datetime
import json
import logging
import urllib

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import SuspiciousOperation
from django.db import DatabaseError
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.template import loader
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views import defaults as default_views
from django.views.decorators.http import require_POST

from hyperkitty.forms import MessageDeleteForm, PostForm, ReplyForm
from hyperkitty.lib.mailman import (
    AddressBannedFromList, ModeratedListException)
from hyperkitty.lib.posting import PostingFailed, post_to_list, reply_subject
from hyperkitty.lib.view_helpers import (
    check_mlist_private, get_months, get_posting_form)
from hyperkitty.models.email import Attachment, Email
from hyperkitty.models.thread import Thread


logger = logging.getLogger(__name__)


@check_mlist_private
def index(request, mlist_fqdn, message_id_hash):
    '''
    Displays a single message identified by its message_id_hash (derived from
    message_id)
    '''
    mlist = request.mlist
    message = get_object_or_404(
        Email, mailinglist=mlist, message_id_hash=message_id_hash)
    if request.user.is_authenticated:
        message.myvote = message.votes.filter(user=request.user).first()
    else:
        message.myvote = None

    # Export button
    export = {
        "url": "%s?message=%s" % (
            reverse("hk_list_export_mbox", kwargs={
                "mlist_fqdn": mlist.name,
                "filename": "%s-%s" % (mlist.name, message.message_id_hash)
                }),
            message.message_id_hash),
        "message": _("Download"),
        "title": _("This message in gzipped mbox format"),
    }

    context = {
        'mlist': mlist,
        'message': message,
        'message_id_hash': message_id_hash,
        'months_list': get_months(mlist),
        'month': message.date,
        'reply_form': get_posting_form(ReplyForm, request, mlist),
        'export': export,
        'posting_enabled': getattr(
            settings, 'HYPERKITTY_ALLOW_WEB_POSTING', True),
        'user_is_owner': mlist.is_owner(request.user),
    }
    return render(request, "hyperkitty/message.html", context)


@check_mlist_private
def attachment(request, mlist_fqdn, message_id_hash, counter, filename):
    """
    Sends the numbered attachment for download. The filename is not used for
    lookup, but validated nonetheless for security reasons.
    """
    mlist = request.mlist
    message = get_object_or_404(
        Email, mailinglist=mlist, message_id_hash=message_id_hash)
    att = get_object_or_404(
        Attachment, email=message, counter=int(counter))
    if att.name != filename:
        raise Http404
    # http://djangosnippets.org/snippets/1710/
    response = HttpResponse(att.get_content())
    response['Content-Type'] = att.content_type
    response['Content-Length'] = att.size
    if att.encoding is not None:
        response['Content-Encoding'] = att.encoding
    # Follow RFC2231, browser support is sufficient nowadays (2012-09)
    response['Content-Disposition'] = 'attachment; filename*=UTF-8\'\'%s' \
        % urllib.parse.quote(att.name.encode('utf-8'))
    return response


@require_POST
@check_mlist_private
def vote(request, mlist_fqdn, message_id_hash):
    """ Vote for or against a given message identified by messageid. """
    if not request.user.is_authenticated:
        return HttpResponse('You must be logged in to vote',
                            content_type="text/plain", status=403)
    mlist = request.mlist
    message = get_object_or_404(
        Email, mailinglist=mlist, message_id_hash=message_id_hash)

    value = int(request.POST['vote'])
    message.vote(value, request.user)

    # Extract all the votes for this message to refresh it
    message.myvote = message.votes.filter(user=request.user).first()
    t = loader.get_template('hyperkitty/fragments/like_form.html')
    html = t.render({
        "mlist": mlist,
        "object": message,
        "message_id_hash": message_id_hash,
    }, request)

    votes = message.get_votes()
    result = {"like": votes["likes"], "dislike": votes["dislikes"],
              "html": html}
    return HttpResponse(json.dumps(result),
                        content_type='application/javascript')


@require_POST
@login_required
@check_mlist_private
def reply(request, mlist_fqdn, message_id_hash):
    """Sends a reply to the list."""
    if not getattr(settings, 'HYPERKITTY_ALLOW_WEB_POSTING', True):
        return HttpResponse('Posting via Hyperkitty is disabled',
                            content_type="text/plain", status=403)
    mlist = request.mlist
    form = get_posting_form(ReplyForm, request, mlist, request.POST)
    if not form.is_valid():
        return HttpResponse(form.errors.as_text(),
                            content_type="text/plain", status=400)
    if form.cleaned_data["newthread"]:
        subject = form.cleaned_data["subject"]
        headers = {}
    else:
        message = get_object_or_404(
            Email, mailinglist=mlist, message_id_hash=message_id_hash)
        subject = reply_subject(message.subject)
        headers = {"In-Reply-To": "<%s>" % message.message_id,
                   "References": "<%s>" % message.message_id, }
    if form.cleaned_data["sender"]:
        headers["From"] = form.cleaned_data["sender"]
    try:
        subscribed_now = post_to_list(
            request, mlist, subject, form.cleaned_data["message"], headers)
    except PostingFailed as e:
        return HttpResponse(str(e), content_type="text/plain", status=500)
    except ModeratedListException as e:
        return HttpResponse(str(e), content_type="text/plain", status=403)
    except AddressBannedFromList as e:
        return default_views.permission_denied(request, e)

    # TODO: if newthread, don't insert the temp mail in the thread, redirect to
    # the new thread. Should we insert the mail in the DB and flag it as
    # "temporary", to be confirmed by a later reception from mailman? This
    # looks complex, because the temp mail should only be visible by its
    # sender.

    if form.cleaned_data["newthread"]:
        html = None
    else:
        email_reply = {
            "sender_name": "%s %s" % (request.user.first_name,
                                      request.user.last_name),
            "sender_address": (form.cleaned_data["sender"]
                               or request.user.email),
            "content": form.cleaned_data["message"],
            # no need to increment, level = thread_depth - 1
            "level": message.thread_depth,
            "mailinglist": mlist,
        }
        t = loader.get_template('hyperkitty/ajax/temp_message.html')
        html = t.render({'email': email_reply}, request)
    # TODO: make the message below translatable.
    result = {"result": _("Your reply has been sent and is being processed."),
              "message_html": html}
    if subscribed_now:
        result['result'] += (
            _("\n  You have been subscribed to {} list.").format(mlist_fqdn))
    return HttpResponse(json.dumps(result),
                        content_type="application/javascript")


@login_required
@check_mlist_private
def new_message(request, mlist_fqdn):
    """Sends a new thread-starting message to the list. """
    if not getattr(settings, 'HYPERKITTY_ALLOW_WEB_POSTING', True):
        return HttpResponse('Posting via Hyperkitty is disabled',
                            content_type="text/plain", status=403)
    mlist = request.mlist
    if request.method == 'POST':
        form = get_posting_form(PostForm, request, mlist, request.POST)
        if form.is_valid():
            today = datetime.date.today()
            headers = {}
            if form.cleaned_data["sender"]:
                headers["From"] = form.cleaned_data["sender"]
            try:
                post_to_list(request, mlist, form.cleaned_data['subject'],
                             form.cleaned_data["message"], headers)
            except PostingFailed as e:
                messages.error(request, str(e))
                logger.info(
                    "Failed to send message from %s to list %s due to %s",
                    form.cleaned_data["sender"], mlist, str(e))
            except ModeratedListException as e:
                return HttpResponse(
                    str(e), content_type="text/plain", status=403)
            except AddressBannedFromList as e:
                return default_views.permission_denied(request, e)
            else:
                messages.success(
                    request, "The message has been sent successfully.")
                redirect_url = reverse('hk_archives_with_month', kwargs={
                    "mlist_fqdn": mlist_fqdn,
                    'year': today.year, 'month': today.month})
                return redirect(redirect_url)
    else:
        form = get_posting_form(PostForm, request, mlist)
    context = {
        "mlist": mlist,
        "post_form": form,
        'months_list': get_months(mlist),
    }
    return render(request, "hyperkitty/message_new.html", context)


@login_required
@check_mlist_private
def delete(request, mlist_fqdn, threadid=None, message_id_hash=None):
    """Delete messages. """
    mlist = request.mlist
    if not (request.user.is_staff or
            request.user.is_superuser or
            mlist.is_owner(request.user)):
        return HttpResponse(
            'You must be a staff member or list owner to delete a message',
            content_type="text/plain", status=403)

    if threadid is not None:
        thread = get_object_or_404(
            Thread, mailinglist=mlist, thread_id=threadid)
        message = None
    elif message_id_hash is not None:
        message = get_object_or_404(
            Email, mailinglist=mlist, message_id_hash=message_id_hash)
        thread = None
    else:
        raise SuspiciousOperation

    form_queryset = Email.objects.filter(mailinglist=mlist)
    if thread is not None:
        form_queryset = form_queryset.filter(thread=thread)
    elif message is not None:
        form_queryset = form_queryset.filter(pk=message.pk)

    if request.method == 'POST':
        form = MessageDeleteForm(request.POST)
        form.fields["email"].queryset = form_queryset
        if form.is_valid():
            thread_ids = []
            for email in sorted(form.cleaned_data["email"], reverse=True):
                email.refresh_from_db()
                thread_id = email.thread.pk
                try:
                    email.delete()
                except DatabaseError as e:
                    form.add_error(
                        "email",
                        _("Could not delete message %(msg_id_hash)s: "
                          "%(error)s")
                        % {"msg_id_hash": email.message_id_hash, "error": e})
                    continue
                logger.info("Deleted email %s (%s)",
                            email.pk, email.message_id)
                thread_ids.append(thread_id)
            if thread_ids:
                messages.success(
                    request, _("Successfully deleted %(count)s messages.")
                    % {"count": len(thread_ids)})
            if not form.has_error("email"):
                if len(set(thread_ids)) == 1:
                    try:
                        thread = Thread.objects.get(pk=thread_ids[0])
                        return redirect(reverse('hk_thread', kwargs={
                            "mlist_fqdn": mlist_fqdn,
                            "threadid": thread.thread_id}))
                    except Thread.DoesNotExist:
                        # The thread has been deleted to in cascade, go back to
                        # the list.
                        pass
                return redirect(reverse('hk_list_overview', kwargs={
                    "mlist_fqdn": mlist_fqdn}))
    else:
        initial = form_queryset.values_list("pk", flat=True)
        form = MessageDeleteForm(initial={"email": initial})
        form.fields["email"].queryset = form_queryset
    context = {
        "mlist": mlist,
        "form": form,
    }
    if thread is not None:
        context.update({
            "thread": thread,
            "form_action": reverse("hk_thread_delete", kwargs={
                "mlist_fqdn": mlist_fqdn, "threadid": thread.thread_id}),
            "cancel_url": reverse("hk_thread", kwargs={
                "mlist_fqdn": mlist_fqdn, "threadid": thread.thread_id}),
            })
    elif message is not None:
        context.update({
            "message": message,
            "form_action": reverse("hk_message_delete", kwargs={
                "mlist_fqdn": mlist_fqdn,
                "message_id_hash": message.message_id_hash}),
            "cancel_url": reverse("hk_message_index", kwargs={
                "mlist_fqdn": mlist_fqdn,
                "message_id_hash": message.message_id_hash}),
            })
    return render(request, "hyperkitty/message_delete.html", context)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import logging

from django.conf import settings
from django.db import models

from allauth.account.models import EmailAddress

from .email import Email


logger = logging.getLogger(__name__)


class Profile(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL,
                                related_name="hyperkitty_profile",
                                on_delete=models.CASCADE)
    karma = models.IntegerField(default=1)

    def __str__(self):
        return '%s' % (str(self.user))

    @property
    def emails(self):
        return Email.objects.filter(
            sender__address__in=self.addresses).order_by("date")

    @property
    def addresses(self):
        return list(EmailAddress.objects.filter(
            user=self.user).filter(
            verified=True).order_by("email").values_list(
            "email", flat=True))

    def get_votes_in_list(self, list_name):
        # TODO: Caching ?
        votes = self.user.votes.filter(email__mailinglist__name=list_name)
        likes = votes.filter(value=1).count()
        dislikes = votes.filter(value=-1).count()
        return likes, dislikes

    def get_first_post(self, mlist):
        return self.emails.filter(
            mailinglist=mlist).order_by("archived_date").first()



# Generated by Django 3.0.13 on 2021-03-14 20:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0021_add_owners_mods'),
    ]

    operations = [
        migrations.AddField(
            model_name='mailinglist',
            name='archive_rendering_mode',
            field=models.CharField(max_length=255, null=True),
        ),
    ]



# Generated by Django 2.2 on 2019-09-07 19:27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0019_auto_20190127_null_description'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='parent',
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name='children',
                to='hyperkitty.Email'),
        ),
    ]



# -*- coding: utf-8 -*-
# flake8: noqa
import django
from django.db import connection, migrations, models


def hotpatch_schema_editor(apps, schema_editor):
    # Patch Django <= 1.8 because it does not know about type
    # conversion in PostgreSQL.
    if django.VERSION < (1, 9) and connection.vendor == "postgresql":
        schema_editor.sql_alter_column_type = \
            "ALTER COLUMN %(column)s TYPE %(type)s USING %(column)s::%(type)s"


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0014_mailinglist_id_2'),
    ]

    operations = [
        migrations.RunPython(hotpatch_schema_editor, hotpatch_schema_editor),
        # Restore the constraints on thread.mailinglist_id and
        # email.mailinglist_id
        migrations.AlterField(
            model_name='thread',
            name='mailinglist',
            field=models.ForeignKey(
                related_name='threads',
                to='hyperkitty.MailingList',
                on_delete=models.CASCADE),
        ),
        migrations.AlterField(
            model_name='email',
            name='mailinglist',
            field=models.ForeignKey(
                related_name='emails',
                to='hyperkitty.MailingList',
                on_delete=models.CASCADE),
        ),
    ]



from contextlib import suppress

from django import template
from django.utils.safestring import mark_safe

from hyperkitty.lib.renderer import markdown_renderer, text_renderer
from hyperkitty.models.mailinglist import ArchiveRenderingMode


register = template.Library()


@register.filter()
def render(email, mlist):
    """Render the email content based on MailingList settings.

    This enables MailingList owners to choose between the two available
    renderer using MailingList settings in Postorius.

    In case the display is rendered in fixed_width font because there is
    a patch in it, do not use markdown mode since that mucks with the
    code.

    :param value: The text value to render.
    :param mlist: The MailingList object this email belongs to.
    :returns: The rendered HTML form the input value text.
    """
    try:
        content = email.content
        display_fixed = email.display_fixed
    except AttributeError:
        content = email.get('content', '')
        display_fixed = email.get('display_fixed', False)

    if (
        mlist.archive_rendering_mode == ArchiveRenderingMode.markdown.name and
        not display_fixed
    ):
        with suppress(TypeError):
            return mark_safe(markdown_renderer(content))
    try:
        return mark_safe(text_renderer(content))
    except (KeyError, ValueError):
        return mark_safe('<pre>' + content + '</pre>')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>

"""
Utilities for management commands.
"""

import logging
import sys


class DualStreamsHandler(logging.StreamHandler):
    """
    Emits messages below and including INFO to the first stream, and messages
    above and including WARNING to the second stream. Useful when the first
    stream is stdout and the second stream is stderr.
    """

    def __init__(self, stream_low=None, stream_high=None):
        super(DualStreamsHandler, self).__init__()
        if stream_low is None:
            stream_low = sys.stdout
        self.stream_low = stream_low
        if stream_high is None:
            stream_high = sys.stderr
        self.stream_high = stream_high

    def emit(self, record):
        """Wraps the original emit method to dynamically switch the stream."""
        self.stream = self.stream_low
        if record.levelno > logging.INFO:
            self.stream = self.stream_high
        super(DualStreamsHandler, self).emit(record)


def setup_logging(command, verbosity):
    """
    Sets up logging to send messages up to INFO level to stdout and messages
    from WARNING level and up to stderr. This is done regardless of the logging
    settings in Django (which may be sending everything to a log file).
    Existing logging settings are not changed.
    """
    if verbosity >= 3:
        debuglevel = logging.DEBUG
    else:
        debuglevel = logging.INFO
    handler = DualStreamsHandler(
        stream_low=command.stdout, stream_high=command.stderr)
    formatter = logging.Formatter(fmt='%(message)s')
    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.setLevel(debuglevel)
    root_logger.addHandler(handler)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import json
import re

from django.conf import settings
from django.db.models import Q
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import redirect, render

from django_mailman3.lib.mailman import get_subscriptions
from django_mailman3.lib.paginator import paginate
from django_mailman3.models import MailDomain

from hyperkitty.models import ArchivePolicy, MailingList


class SortMode:
    """All the Sort modes the Index page supports."""
    CREATION = "creation"
    POPULAR = "popular"
    NAME = "name"
    ACTIVE = "active"

    DEFAULT = POPULAR

    @classmethod
    def all(self):
        """Return all the Sort modes."""
        return (self.NAME, self.ACTIVE, self.POPULAR, self.CREATION)


def index(request):
    mlists = MailingList.objects.all()
    sort_mode = request.GET.get("sort", SortMode.DEFAULT)

    if sort_mode not in SortMode.all():
        return HttpResponseBadRequest("Wrong sort mode parameter")

    # Domain filtering
    if getattr(settings, 'FILTER_VHOST', False):
        our_lists = MailingList.objects.none()
        domain = request.get_host().split(":")[0]
        mail_hosts = []
        for mlist in mlists:
            mail_host = re.sub('^.*@', '', mlist.name)
            try:
                if (MailDomain.objects.get(
                        mail_domain=mail_host).site.domain == domain):
                    if mail_host not in mail_hosts:
                        mail_hosts.append(mail_host)
            except MailDomain.DoesNotExist:
                pass
        if len(mail_hosts) == 0:
            mail_hosts = [domain]
        for domain in mail_hosts:
            domain = '@%s' % domain
            our_lists = our_lists | mlists.filter(name__iendswith=domain)
        mlists = our_lists

    # Name filtering
    name_filter = request.GET.get('name')
    if name_filter:
        sort_mode = SortMode.NAME
        mlists = mlists.filter(
            Q(name__icontains=name_filter) |
            Q(display_name__icontains=name_filter)
            )
        if mlists.count() == 1:
            return redirect(
                'hk_list_overview', mlist_fqdn=mlists.first().name)

    # Access Filtering
    if request.user.is_superuser:
        # Superusers see everything
        pass
    elif request.user.is_authenticated:
        # For authenticated users show only their subscriptions
        # and public lists
        mlists = mlists.filter(
            Q(list_id__in=get_subscriptions(request.user)) |
            Q(archive_policy=ArchivePolicy.public.value)
        )
    else:
        # Unauthenticated users only see public lists
        mlists = mlists.filter(
            archive_policy=ArchivePolicy.public.value)

    # Sorting
    if sort_mode == SortMode.NAME:
        mlists = mlists.order_by(SortMode.NAME)
    elif sort_mode == SortMode.ACTIVE:
        mlists = list(mlists)
        mlists.sort(key=lambda lk: lk.recent_threads_count, reverse=True)
    elif sort_mode == SortMode.POPULAR:
        mlists = list(mlists)
        mlists.sort(key=lambda lk: lk.recent_participants_count, reverse=True)
    elif sort_mode == SortMode.CREATION:
        mlists = mlists.order_by("-created_at")

    # Inactive List Setting
    show_inactive = getattr(settings, 'SHOW_INACTIVE_LISTS_DEFAULT', False)

    mlists = paginate(mlists, request.GET.get('page'),
                      request.GET.get('count'))

    context = {
        'view_name': 'all_lists',
        'all_lists': mlists,
        'sort_mode': sort_mode,
        'show_inactive': show_inactive
    }
    return render(request, "hyperkitty/index.html", context)


def find_list(request):
    """Get a list of mailing list matching the query parameter 'term'.

    This serves the API that the list search uses, powered by jquery's
    autocomplete.

    The default returned content_type is `application/javascript`.
    """
    term = request.GET.get('term')
    result = []
    if term:
        query = MailingList.objects.filter(
            Q(name__icontains=term) | Q(display_name__icontains=term)
            )
        if not request.user.is_superuser:
            query = query.filter(
                archive_policy=ArchivePolicy.public.value)
        query = query.order_by("name").values("name", "display_name")
        for line in query[:20]:
            result.append({
                "value": line["name"],
                "label": line["display_name"] or line["name"],
            })
    return HttpResponse(
        json.dumps(result), content_type='application/javascript')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#
import datetime
import logging
from enum import Enum
from urllib.error import HTTPError

from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.utils.timezone import now, utc

import dateutil.parser
from django_mailman3.lib.mailman import get_mailman_client
from django_mailman3.lib.user import get_django_user
from mailmanclient import MailmanConnectionError

from hyperkitty.lib.utils import pgsql_disable_indexscan
from .common import ModelCachedValue
from .thread import Thread


logger = logging.getLogger(__name__)


class ArchivePolicy(Enum):
    """
    Copy from mailman.interfaces.archiver.ArchivePolicy since we can't import
    mailman (PY3-only).

    This should probably be moved to mailman.client.
    """
    never = 0
    private = 1
    public = 2


class ArchiveRenderingMode(Enum):
    """Archive rendering mode determines if Hyperkitty renders archive as plain
    text or markdown text.
    """
    text = 'text'
    markdown = 'markdown'


class MailingList(models.Model):
    """
    An archived mailing-list.
    """
    name = models.CharField(db_index=True, max_length=254, unique=True)
    list_id = models.CharField(max_length=254, null=True, unique=True)
    display_name = models.CharField(max_length=255)
    description = models.TextField(null=True)
    subject_prefix = models.CharField(max_length=255)
    archive_policy = models.IntegerField(
        choices=[(p.value, p.name) for p in ArchivePolicy],
        default=ArchivePolicy.public.value)
    created_at = models.DateTimeField(default=now)
    archive_rendering_mode = models.CharField(max_length=255, null=True)

    owners = models.ManyToManyField(settings.AUTH_USER_MODEL,
                                    related_name='owned_lists')

    moderators = models.ManyToManyField(settings.AUTH_USER_MODEL,
                                        related_name='moderated_lists')

    MAILMAN_ATTRIBUTES = (
        "display_name", "description", "subject_prefix",
        "archive_policy", "created_at", "list_id", "archive_rendering_mode",
    )

    def __init__(self, *args, **kwargs):
        super(MailingList, self).__init__(*args, **kwargs)
        self.cached_values = {
            "recent_threads": RecentThreads(self),
            "participants_count_for_month": ParticipantsCountForMonth(self),
            "recent_participants_count": RecentParticipantsCount(self),
            "top_posters": TopPosters(self),
            "top_threads": TopThreads(self),
            "popular_threads": PopularThreads(self),
            "first_date": FirstDate(self),
        }
        self.recent_cached_values = [
            self.cached_values[key] for key in [
                "recent_threads",
                "recent_participants_count",
                "top_threads", "top_posters", "popular_threads",
                ]
            ]

    def __repr__(self):
        return '<MailingList {} ({})>'.format(self.name, self.list_id)

    def __str__(self):
        return self.list_id

    @property
    def is_private(self):
        return self.archive_policy == ArchivePolicy.private.value

    @property
    def is_new(self):
        return (self.created_at and
                now() - self.created_at <= datetime.timedelta(days=30))

    def get_recent_dates(self):
        today = now()
        # today -= datetime.timedelta(days=400) #debug
        # the upper boundary is excluded in the search, add one day
        end_date = today + datetime.timedelta(days=1)
        begin_date = end_date - datetime.timedelta(days=32)
        return begin_date, end_date

    def get_participants_count_between(self, begin_date, end_date):
        # We filter on emails dates instead of threads dates because that would
        # also include last month's participants when threads carry from one
        # month to the next
        # TODO: caching?
        return self.emails.filter(
                date__gte=begin_date, date__lt=end_date
            ).values("sender_id").distinct().count()

    def get_threads_between(self, begin_date, end_date):
        """Get MailingList threads between the dates.

        :param begin_date: The cutoff for oldest thread, decided by the posting
            date of the first email.
        :param end_date: The cutoff for the newest thread.
        """
        return (self.threads.filter(
            starting_email__date__lt=end_date, date_active__gte=begin_date)
                .order_by("-date_active")
                .select_related('starting_email')
                .select_related('starting_email__sender')
                )

    @property
    def recent_participants_count(self):
        return self.cached_values["recent_participants_count"]()

    @property
    def recent_threads(self):
        return self.cached_values["recent_threads"]()

    @property
    def recent_threads_count(self):
        # Don't use a CachedValue for this one, because it does not need a
        # specific warm up or rebuild: this is done by the recent_threads
        # CachedValue.
        begin_date, end_date = self.get_recent_dates()
        cache_key = "MailingList:%s:recent_threads_count" % self.pk
        result = cache.get(cache_key)
        if result is None:
            result = self.get_threads_between(begin_date, end_date).count()
            # The cache will be refreshed daily by a periodic job.
            cache.set(cache_key, result, None)
        return result

    def get_participants_count_for_month(self, year, month):
        return self.cached_values["participants_count_for_month"](year, month)

    @property
    def top_posters(self):
        return self.cached_values["top_posters"]()

    @property
    def top_threads(self):
        """Threads with the most answers."""
        return self.cached_values["top_threads"]()

    @property
    def popular_threads(self):
        """Threads with the most votes."""
        return self.cached_values["popular_threads"]()

    def update_from_mailman(self):
        try:
            client = get_mailman_client()
            mm_list = client.get_list(self.name)
        except MailmanConnectionError:
            return
        except HTTPError:
            return  # can't update at this time
        if not mm_list:
            return

        def convert_date(value):
            value = dateutil.parser.parse(value)
            if value.tzinfo is None:
                value = value.replace(tzinfo=utc)
            return value
        converters = {
            "created_at": convert_date,
            "archive_policy": lambda p: ArchivePolicy[p].value,
        }
        for propname in self.MAILMAN_ATTRIBUTES:
            try:
                value = getattr(mm_list, propname)
            except AttributeError:
                value = mm_list.settings[propname]
            if propname in converters:
                value = converters[propname](value)
            setattr(self, propname, value)
        self.save()
        # Update the owners and moderators from Mailman 3 API.
        self._update_owner_mods_from_mailman(mm_list)

    def _update_owner_mods_from_mailman(self, mm_list):
        """Update Owners and moderators from Mailman."""
        for owner in mm_list.owners:
            user = get_django_user(owner)
            if user and not bool(self.owners.filter(id=user.id)):
                logger.debug('Adding %s as owner of MailingList: %s',
                             user, self.list_id)
                self.owners.add(user)

        for mod in mm_list.moderators:
            user = get_django_user(mod)
            if user and not bool(self.moderators.filter(id=user.id)):
                logger.debug('Adding %s as moderator of MailingList: %s',
                             user, self.list_id)
                self.moderators.add(user)

    def is_owner(self, user):
        try:
            client = get_mailman_client()
            mm_list = client.get_list(self.name)
        except MailmanConnectionError:
            return False
        except HTTPError:
            return False
        if not mm_list:
            return False
        if user.is_authenticated and user.email in [
                owner.address.email for owner in mm_list.owners]:
            return True
        return False

    # Events (signal callbacks)

    def on_pre_save(self):
        # Set the default list_id
        if self.list_id is None:
            self.list_id = self.name.replace("@", ".")

    def on_thread_added(self, thread):
        self.cached_values["recent_threads"].add_thread(thread)

    def on_thread_deleted(self, thread):
        from hyperkitty.tasks import (
            rebuild_mailinglist_cache_for_month,
            rebuild_mailinglist_cache_recent)
        begin_date, end_date = self.get_recent_dates()
        if thread.date_active >= begin_date and thread.date_active < end_date:
            # It's a recent thread
            rebuild_mailinglist_cache_recent(self.name)
        rebuild_mailinglist_cache_for_month(
            self.name, thread.date_active.year, thread.date_active.month)

    def on_email_added(self, email):
        if getattr(settings, "HYPERKITTY_BATCH_MODE", False):
            # Cache handling will be done at the end of the import
            # process.
            return
        # Rebuild the cached values.
        from hyperkitty.tasks import (
            rebuild_mailinglist_cache_for_month,
            rebuild_mailinglist_cache_recent)
        rebuild_mailinglist_cache_recent(self.name)
        rebuild_mailinglist_cache_for_month(
            self.name, email.date.year, email.date.month)

    def on_email_deleted(self, email):
        # Don't use on_email_added, it will try appending to the
        # recent_threads and emails aren't associated to a thread
        # when they are deleted.
        # It's not semantically identical to on_thread_deleted() but it's the
        # same code, so DRY.
        try:
            email.thread
        except Thread.DoesNotExist:
            pass  # Already deleted
        else:
            self.on_thread_deleted(email.thread)

    def on_vote_added(self, vote):
        from hyperkitty.tasks import rebuild_cache_popular_threads
        rebuild_cache_popular_threads(self.name)

    on_vote_deleted = on_vote_added


class RecentThreads(ModelCachedValue):

    cache_key = "recent_threads"

    def get_value(self):
        # Only cache the list of thread ids, or it may go over memcached's size
        # limit (1MB)
        begin_date, end_date = self.instance.get_recent_dates()
        thread_ids = self.instance.get_threads_between(
            begin_date, end_date).values_list("id", flat=True)
        # Convert QuerySet to a list, so that we can add/remove thread_ids.
        return list(thread_ids)

    def rebuild(self):
        value = super(RecentThreads, self).rebuild()
        cache.set("%s_count" % self._get_cache_key(), len(value), None)
        return value

    def get_or_set(self):
        thread_ids = super(RecentThreads, self).get_or_set()
        return [Thread.objects.get(pk=pk) for pk in thread_ids]

    def add_thread(self, thread):
        # Add the thread to the recent_threads.
        # Just append to the cache, a daily cron job will rebuild
        # the cache entirely to remove older threads.
        recent_thread_ids = self.get_value()
        if recent_thread_ids is None:
            recent_thread_ids = []
        if thread.id in recent_thread_ids:
            # If the thread is already recent, make it the most recent.
            recent_thread_ids.remove(thread.id)
        recent_thread_ids.insert(0, thread.id)
        cache.set(self._get_cache_key(), recent_thread_ids, None)
        cache.set("%s_count" % self._get_cache_key(),
                  len(recent_thread_ids), None)


class ParticipantsCountForMonth(ModelCachedValue):

    def _get_cache_key(self, year, month):
        return "MailingList:%s:p_count_for:%s:%s" % (
            self.instance.pk, year, month)

    def get_value(self, year, month):
        begin_date = datetime.datetime(year, month, 1, tzinfo=utc)
        end_date = begin_date + datetime.timedelta(days=32)
        end_date = end_date.replace(day=1)
        return self.instance.get_participants_count_between(
            begin_date, end_date)


class RecentParticipantsCount(ModelCachedValue):

    cache_key = "recent_participants_count"

    def get_value(self):
        begin_date, end_date = self.instance.get_recent_dates()
        return self.instance.get_participants_count_between(
            begin_date, end_date)


class TopPosters(ModelCachedValue):

    cache_key = "top_posters"

    def get_value(self):
        from .email import Email  # avoid circular imports
        begin_date, end_date = self.instance.get_recent_dates()
        query = Email.objects.filter(
                mailinglist=self.instance,
                date__gte=begin_date,
                date__lt=end_date,
            ).only("sender", "sender_name").select_related("sender")
        posters = {}
        for email in query:
            key = (email.sender.address, email.sender_name)
            if key not in posters:
                posters[key] = 1
            else:
                posters[key] += 1
        # It's not necessary to return instances since it's only used in
        # templates where access to instance attributes or dictionnary keys is
        # identical.
        posters = [
            {"address": p[0], "name": p[1], "count": c}
            for p, c in posters.items()
            ]
        sorted_posters = sorted(
            posters, key=lambda p: p["count"], reverse=True)
        return sorted_posters[:5]


class TopThreads(ModelCachedValue):
    """Threads with the most answers."""

    cache_key = "top_threads"

    def get_value(self):
        # Filter on the recent_threads ids instead of re-using the date
        # filter, otherwise the Sum will be computed for every thread
        # regardless of their date.
        begin_date, end_date = self.instance.get_recent_dates()
        recent_thread_ids = self.instance.get_threads_between(
            begin_date, end_date).values("id")
        threads = Thread.objects.filter(
            id__in=recent_thread_ids).annotate(
            models.Count("emails")).order_by("-emails__count")[:20]
        # (not sure about using .values_list() here because of the annotation)
        # Only cache the list of thread ids, or it may go over memcached's size
        # limit (1MB)
        return [t.id for t in threads]

    def get_or_set(self):
        thread_ids = super(TopThreads, self).get_or_set()
        return [Thread.objects.get(pk=pk) for pk in thread_ids]


class PopularThreads(ModelCachedValue):
    """Threads with the most votes."""

    cache_key = "popular_threads"

    def get_value(self):
        # Filter on the recent_threads ids instead of re-using the date
        # filter, otherwise the Sum will be computed for every thread
        # regardless of their date.
        begin_date, end_date = self.instance.get_recent_dates()
        recent_thread_ids = self.instance.get_threads_between(
            begin_date, end_date).values("id")
        threads = Thread.objects.filter(
            id__in=recent_thread_ids).annotate(
            models.Sum("emails__votes__value")).order_by(
            "-emails__votes__value__sum")[:20]
        # (not sure about using .values_list() here because of the annotation)
        for thread in threads:
            value = thread.emails__votes__value__sum
            if value is None:
                value = 0
            cache.set("Thread:%s:votes_total" % thread.id, value, None)
        # Only cache the list of thread ids, or it may go over memcached's size
        # limit (1MB)
        return [t.id for t in threads if t.votes_total > 0]

    def get_or_set(self):
        thread_ids = super(PopularThreads, self).get_or_set()
        return [Thread.objects.get(pk=pk) for pk in thread_ids]


class FirstDate(ModelCachedValue):

    cache_key = "first_date"

    def get_value(self):
        with pgsql_disable_indexscan():
            value = self.instance.emails.order_by(
                "date").values_list("date", flat=True).first()
        if value is not None:
            return value.date()
        else:
            return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>

"""
Import the content of a mbox file into the database.
"""

import mailbox
import os
import re
from contextlib import suppress
from datetime import datetime
from email import message_from_bytes, policy
from email.utils import make_msgid, unquote
from math import floor
from traceback import print_exc

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
from django.db import Error as DatabaseError
from django.db import transaction
from django.utils.formats import date_format
from django.utils.timezone import utc

from dateutil import tz
from dateutil.parser import parse as parse_date

from hyperkitty.lib.analysis import compute_thread_order_and_depth
from hyperkitty.lib.incoming import DuplicateMessage, add_to_list
from hyperkitty.lib.mailman import sync_with_mailman
from hyperkitty.lib.utils import get_message_id
from hyperkitty.management.utils import setup_logging
from hyperkitty.models import Email, MailingList, Thread


# Allow all wierd line endings.
TEXTWRAP_RE = re.compile(r"(\n|\r|\r\n|\n\r)\s*")


class ProgressMarker(object):

    def __init__(self, verbose, stdout):
        self.verbose = verbose
        self.total = None
        self.count = 0
        self.count_imported = 0
        self.spinner_seq = ('|', '/', '-', '\\')
        self.stdout = stdout

    def tick(self, msgid=None):
        if self.total:
            msg = "%d%%" % floor(100.0 * self.count / self.total)
        else:
            msg = self.spinner_seq[self.count % len(self.spinner_seq)]
        if self.verbose:
            if self.total:
                self.stdout.write(
                    "%s (%d/%d, %s)" % (msgid, self.count, self.total, msg))
            else:
                self.stdout.write("%s (%d)" % (msgid, self.count))
        else:
            self.stdout.write("\r%s" % msg, ending='')
            self.stdout.flush()
        self.count += 1

    def finish(self):
        if self.verbose:
            self.stdout.write('  %s emails read' % self.count)
            self.stdout.write('  %s email added to the database'
                              % self.count_imported)
        else:
            self.stdout.write("\r", ending='')
            self.stdout.flush()


class DbImporter(object):
    """
    Import email messages into the HyperKitty database using its API.
    """

    def __init__(self, list_address, options, stdout, stderr):
        self.list_address = list_address
        self.verbose = options["verbosity"] >= 2
        self.since = options.get("since")
        self.impacted_thread_ids = set()
        self.stdout = stdout
        self.stderr = stderr

    def _is_too_old(self, message, report_name):
        if not self.since:
            return False
        date = message.get("date")
        if not date:
            return False
        try:
            date = parse_date(date)
        except ValueError as e:
            if self.verbose:
                self.stderr.write(
                    "Can't parse date string in message {}{}: {}. "
                    "The date string is: '{}'".format(
                        unquote(message.get("message-id", 'n/a')),
                        report_name, e, date.decode("ascii", "replace")))
            return False
        if date.tzinfo is None:
            date = date.replace(tzinfo=utc)
        try:
            return date <= self.since
        except ValueError:
            return False

    def _get_date(self, message, header, report_name):
        try:
            date = message.get(header)
            if date and not date.datetime:
                if self.verbose:
                    self.stderr.write(
                        "Bad datetime in {} header in message {}{}.".format(
                            header, unquote(message.get("message-id", 'n/a')),
                            report_name))
                return None
        except (AttributeError, TypeError, ValueError) as e:
            if self.verbose:
                self.stderr.write(
                    "Can't get {} header in message {}{}: {}.".format(
                        header, unquote(message.get("message-id", 'n/a')),
                        report_name, e))
            return None
        return date

    def _fix_mid(self, mid):
        # Message-IDs of the form
        # Message-ID: <58C6B8A302CF0E71@example.com> (added by
        #  postmaster@example.com)
        # have been seen.  First remove whitespace. Then if the Message-ID
        # starts with '<' remove everything after the first '>'. Then ensure
        # it has '<' and '>' at beginning and end.
        mid = re.sub(r'\s', '', mid)
        if mid.startswith('<'):
            mid = re.sub('>.*', '>', mid)
        else:
            mid = '<' + mid
        if not mid.endswith('>'):
            mid += '>'
        # We've also seen bogus Message-IDs with non-ascii so remove any.
        mid = mid.encode('ascii', 'ignore').decode('ascii')
        return mid

    def from_mbox(self, mbfile, report_name):
        """
        Insert all the emails contained in an mbox file into the database.

        :arg mbfile: a mailbox file
        """
        mbox = mailbox.mbox(mbfile)
        progress_marker = ProgressMarker(self.verbose, self.stdout)
        if not self.since:
            progress_marker.total = len(mbox)
        for msg in mbox:
            # FIXME: this converts mailbox.mboxMessage to
            # email.message.EmailMessage
            # We need to fix up Message-IDs here because recent Python email
            # throws an IndexError exception on getting a Message-ID of <>.
            if msg['message-id'] is None:
                msg['Message-ID'] = make_msgid('generated')
            mid = str(msg['message-id'])
            if mid.strip() in ('', '<>'):
                msg.replace_header('Message-ID', make_msgid('generated'))
            elif mid != self._fix_mid(mid):
                msg.replace_header('Message-ID', self._fix_mid(mid))
            try:
                msg_raw = msg.as_bytes(unixfrom=False)
            except UnicodeError as e:
                self.stderr.write('Failed to convert {}{} to bytes\n'
                                  '    {}'.format(
                                   unquote(msg.get("Message-Id", 'n/a')),
                                   report_name, e))
                continue
            unixfrom = msg.get_from()
            try:
                message = message_from_bytes(msg_raw, policy=policy.default)
            except (UnicodeError, IndexError) as e:
                self.stderr.write('Failed to convert {}{} to '
                                  'email.message.Message\n    {}'.format(
                                   unquote(msg.get("Message-Id", 'n/a')),
                                   report_name, e))
                continue
            # Fix missing and wierd Date: headers.
            date = (self._get_date(message, "date", report_name) or
                    self._get_date(message, "resent-date", report_name))
            if unixfrom and not date:
                date = " ".join(unixfrom.split()[1:])

            if date:
                # Make sure this date can be parsed before setting it as as the
                # header. If not, a TypeError is raised and we just keep the
                # old Header.
                with suppress(TypeError):
                    del message['Date']
                    message['Date'] = date

            if self._is_too_old(message, report_name):
                continue
            progress_marker.tick(unquote(message.get("message-id", 'n/a')))
            # Un-wrap the subject line if necessary
            if message["subject"]:
                # If we can't replace the header because it contains some
                # unicode Next Line or similar. Just keep the original.
                with suppress(ValueError):
                    message.replace_header(
                        "subject", TEXTWRAP_RE.sub(" ", message["subject"]))
            if unixfrom:
                message.set_unixfrom(unixfrom)
            # Now insert the message
            try:
                with transaction.atomic():
                    add_to_list(self.list_address, message, from_import=True)
            except DuplicateMessage as e:
                if self.verbose:
                    self.stderr.write(
                        "Duplicate email with message-id '%s'%s"
                        % (e.args[0], report_name))
                continue
            except (LookupError, UnicodeError, ValueError) as e:
                self.stderr.write("Failed adding message %s%s: %s"
                                  % (message.get("Message-ID"),
                                     report_name, e))
                if len(e.args) == 2:
                    try:
                        self.stderr.write(
                            "%s from %s about %s"
                            % (e.args[0], e.args[1].get("From"),
                               e.args[1].get("Subject")))
                    except UnicodeDecodeError:
                        pass
                # Don't reraise the exception
                continue
            except DatabaseError:
                try:
                    print_exc(file=self.stderr)
                except UnicodeError:
                    pass
                self.stderr.write(
                    "Message %s%s failed to import, skipping"
                    % (unquote(message["Message-Id"]), report_name))
                continue
            except Exception as e:
                # In case of *any* exception, log and continue to import the
                # rest of the archive.
                self.stderr.write(
                    "Message {}{} failed to import, skipping\n    {}".format(
                        unquote(message["Message-ID"]), report_name, e))
                continue
            email = Email.objects.get(
                mailinglist__name=self.list_address,
                message_id=get_message_id(message))
            # # Commit every time to be able to rollback on error
            # if not transaction.get_autocommit():
            #     transaction.commit()
            # Store the list of impacted threads to be able to compute the
            # thread_order and thread_depth values
            self.impacted_thread_ids.add(email.thread_id)
            progress_marker.count_imported += 1
        # self.store.search_index.flush() # Now commit to the search index
        progress_marker.finish()
        mbox.close()


class Command(BaseCommand):
    help = (
        "Imports the specified archive mbox(es). "
        "Before running this, the mbox(es) should be checked for "
        "messages that could throw uncaught exceptions with the "
        "contrib/check_hk_archive script and for unescaped 'From ' lines "
        "with Mailman 2.1's cleanarch script.")

    def add_arguments(self, parser):
        parser.add_argument('mbox', nargs='+')
        parser.add_argument(
            '--delete',
            action='store_true',
            dest='delete',
            default=False,
            help='Delete poll instead of closing it')
        parser.add_argument(
            '-l', '--list-address',
            help="the full list address the mailbox will be imported to")
        parser.add_argument(
            '--no-sync-mailman',
            action='store_true', default=False,
            help="do not sync properties with Mailman (faster, useful "
                 "for batch imports)")
        parser.add_argument(
            '--since',
            help="only import emails later than this date.  Defaults to the "
                 "date of the newest message in the existing archive if any.")
        parser.add_argument(
            '--ignore-mtime',
            action='store_true', default=False,
            help="do not check mbox mtimes (slower)")

    def _check_options(self, options):
        if not options.get("list_address"):
            raise CommandError(
                "The list address must be given on the command-line.")
        if "@" not in options["list_address"]:
            raise CommandError(
                "The list address must be fully-qualified, including "
                "the '@' symbol and the domain name.")
        if not options.get("mbox"):
            raise CommandError("No mbox file selected.")
        for mbfile in options["mbox"]:
            if not os.path.exists(mbfile):
                raise CommandError("No such file: %s" % mbfile)
        options["verbosity"] = int(options.get("verbosity", "1"))
        if options["since"]:
            try:
                options["since"] = parse_date(options["since"])
                if options["since"].tzinfo is None:
                    options["since"] = options["since"].replace(
                        tzinfo=tz.tzlocal())
            except ValueError as e:
                raise CommandError("invalid value for '--since': %s" % e)

    def handle(self, *args, **options):
        self._check_options(options)
        setup_logging(self, options["verbosity"])
        # main
        list_address = options["list_address"].lower()
        # Keep autocommit on SQLite:
        # https://docs.djangoproject.com/en/1.8/topics/db/transactions/#savepoints-in-sqlite
        # if (settings.DATABASES["default"]["ENGINE"]
        #     != "django.db.backends.sqlite3":
        #     transaction.set_autocommit(False)
        # Sync list settings with Mailman before importing messages:
        if not options["no_sync_mailman"]:
            mlist = MailingList.objects.get_or_create(name=list_address)[0]
            mlist.update_from_mailman()
            mlist.save()
        settings.HYPERKITTY_BATCH_MODE = True
        # Only import emails newer than the latest email in the DB
        latest_email_date = Email.objects.filter(
                mailinglist__name=list_address
            ).values("date").order_by("-date").first()
        if latest_email_date and not options["since"]:
            options["since"] = latest_email_date["date"]
            self.stdout.write(
                'Warning: not importing messages older than {}. '
                'Use "--since=<date>" to import older messages.'.format(
                    date_format(latest_email_date["date"])))
        if options["since"] and options["verbosity"] >= 2:
            self.stdout.write(
                "Only emails after %s will be imported" % options["since"])
        importer = DbImporter(list_address, options, self.stdout, self.stderr)
        # disable mailman client for now
        for mbfile in options["mbox"]:
            if len(options["mbox"]) > 1:
                report_name = ' from mbox {}'.format(mbfile)
            else:
                report_name = ''
            if options["verbosity"] >= 1:
                self.stdout.write("Importing from mbox file %s to %s"
                                  % (mbfile, list_address))
            if not options["ignore_mtime"] and options["since"] is not None:
                mtime = datetime.fromtimestamp(
                    os.path.getmtime(mbfile), tz.tzlocal())
                if mtime <= options["since"]:
                    if options["verbosity"] >= 2:
                        self.stdout.write('Mailbox file for %s is too old'
                                          % list_address)
                    continue
            importer.from_mbox(mbfile, report_name)
            if options["verbosity"] >= 2:
                total_in_list = Email.objects.filter(
                    mailinglist__name=list_address).count()
                self.stdout.write('  %s emails are stored into the database'
                                  % total_in_list)
        if options["verbosity"] >= 1:
            self.stdout.write("Computing thread structure")
        # Work on batches of thread ids to avoid creating a huge SQL request
        # (it's an IN statement)
        thread_ids = list(importer.impacted_thread_ids)
        while thread_ids:
            thread_ids_batch = thread_ids[:100]
            thread_ids = thread_ids[100:]
            for thread in Thread.objects.filter(id__in=thread_ids_batch):
                compute_thread_order_and_depth(thread)
        if not options["no_sync_mailman"]:
            if options["verbosity"] >= 1:
                self.stdout.write("Synchronizing properties with Mailman")
            sync_with_mailman()
            # if not transaction.get_autocommit():
            #     transaction.commit()
        if options["verbosity"] >= 1:
            self.stdout.write("Warming up cache")
        call_command("hyperkitty_warm_up_cache", list_address)
        if options["verbosity"] >= 1:
            self.stdout.write(
                "The full-text search index is not updated for this list. "
                "It will not be updated by the 'minutely' incremental "
                "update job. To update the index for this list, run the "
                "Django admin command with arguments "
                "'update_index_one_list {}'."
                .format(list_address)
                )



# -*- coding: utf-8 -*-
# Copyright (C) 2020-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Stasiek Michalski <stasiek@michalski.cc>
#
import re

from django.conf import settings
from django.contrib.syndication.views import Feed
from django.shortcuts import get_object_or_404
from django.urls import reverse

from hyperkitty.models import Email, MailingList


def chr2entity(mo):
    return '&#' + str(ord(mo.group(0))) + ';'


def sanitize(x):
    if isinstance(x, str):
        return re.sub(r'[\x00-\x08\x0b\x0c\x0e-\x1f\x7f-\xa0]',
                      chr2entity, x)
    else:
        return x


class MailingListFeed(Feed):

    def get_object(self, request, mlist_fqdn):
        return get_object_or_404(MailingList, name=mlist_fqdn)

    def title(self, obj):
        return sanitize(obj.display_name)

    def link(self, obj):
        return reverse("hk_list_overview", kwargs={"mlist_fqdn": obj.name})

    def description(self, obj):
        return sanitize(obj.description)

    def items(self, obj):
        len = getattr(settings, 'HYPERKITTY_MLIST_FEED_LENGTH', 30)
        return Email.objects.filter(mailinglist=obj).order_by('-date')[:len]

    def item_title(self, item):
        return sanitize(item.subject)

    def item_link(self, item):
        return reverse("hk_message_index",
                       kwargs={"mlist_fqdn": item.mailinglist.name,
                               "message_id_hash": item.message_id_hash})

    def item_description(self, item):
        return sanitize(item.content.replace("@", "\uff20"))

    def item_author_name(self, item):
        return sanitize(item.sender.name)

    def item_author_email(self, item):
        return sanitize(item.sender.address.replace("@", "\uff20"))

    def item_author_link(self, item):
        return reverse("hk_public_user_profile",
                       kwargs={"user_id": item.sender.mailman_id})

    def item_pubdate(self, item):
        return sanitize(item.date)



# -*- coding: utf-8 -*-
# flake8: noqa

from django.db import connection, migrations, models


def populate_sender_name(apps, schema_editor):
    #Email = apps.get_model("hyperkitty", "Email")
    #for email in Email.objects.only("sender").select_related("sender"):
    #    email.sender_name = email.sender.name
    #    email.save()
    # Don't use the model, use a single UPDATE query, it's much faster.
    cursor = connection.cursor()
    cursor.execute("""
        UPDATE hyperkitty_email SET sender_name = (
            SELECT name FROM hyperkitty_sender
            WHERE address = hyperkitty_email.sender_id LIMIT 1
        )
        """)

def populate_sender_name_reverse(apps, schema_editor):
    Sender = apps.get_model("hyperkitty", "Sender")
    for sender in Sender.objects.all():
        for email_sender_name in sender.emails.order_by(
                "-date").values_list("sender_name", flat=True):
            if email_sender_name:
                sender.name = email_sender_name
                sender.save()
                break


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0009_duplicate_persona_users'),
    ]

    operations = [
        migrations.AddField(
            model_name='email',
            name='sender_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.RunPython(
            populate_sender_name, populate_sender_name_reverse),
        migrations.RemoveField(
            model_name='sender',
            name='name',
        ),
    ]



# Generated by Django 2.1a1 on 2018-05-30 16:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0016_auto_20180309_0056'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attachment',
            name='content',
            field=models.BinaryField(null=True),
        ),
        migrations.AlterField(
            model_name='attachment',
            name='size',
            field=models.IntegerField(null=True),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import hmac
import json
import logging
from email import message_from_binary_file
from email.message import EmailMessage
from email.policy import default
from functools import wraps
from urllib.parse import unquote, urljoin

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.http import HttpResponse
from django.urls import reverse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from django_mailman3.models import MailDomain

from hyperkitty.lib.incoming import DuplicateMessage, add_to_list
from hyperkitty.lib.utils import get_message_id_hash


logger = logging.getLogger(__name__)


def key_and_ip_auth(func):
    @wraps(func)
    def _decorator(request, *args, **kwargs):
        for attr in ('MAILMAN_ARCHIVER_KEY', 'MAILMAN_ARCHIVER_FROM'):
            if not hasattr(settings, attr):
                msg = "Missing setting: %s" % attr
                logger.error(msg)
                raise ImproperlyConfigured(msg)
        # '*' matches any host
        if '*' in settings.MAILMAN_ARCHIVER_FROM:
            logger.debug(
                "Found '*' into MAILMAN_ARCHIVER_FROM. IP check disabled "
                "for archiving API endpoint. Accepting client IP "
                "{}.".format(request.META["REMOTE_ADDR"]))
        elif (request.META.get("REMOTE_ADDR") not in
                settings.MAILMAN_ARCHIVER_FROM):
            logger.error(
                "Access to the archiving API endpoint was forbidden from "
                "IP {}, your MAILMAN_ARCHIVER_FROM setting may be "
                "misconfigured".format(request.META["REMOTE_ADDR"]))
            return HttpResponse(
                """<html><title>Forbidden</title><body>
                <h1>Access is forbidden</h1><p>Please check the IP addresses
                 assigned to MAILMAN_ARCHIVER_FROM in the settings file.
                </p></body></html>""",
                content_type="text/html", status=403)

        authorization = request.headers.get('Authorization')
        if authorization is None:
            if request.GET.get('key') is not None:
                # Old authentication method used, need to upgrade
                # mailman-hyperkitty
                logger.error(
                    'The MAILMAN_ARCHIVER_KEY is now required to be '
                    'sent over the Authorization HTTP header, you need to '
                    'upgrade the mailman-hyperkitty package to 1.2.0 or newer.'
                )
                return HttpResponse(
                    """<html><title>Auth required</title><body>
                    <h1>Authorization Required</h1><p>The archiver key is now
                     required to be sent over the Authorization HTTP header.
                     You need to upgrade the mailman-hyperkitty package to
                     1.2.0 or newer.
                    </p></body></html>""",
                    content_type="text/html", status=401)
            else:
                # No auth provided over header nor GET param
                logger.error(
                    'The MAILMAN_ARCHIVER_KEY was not sent as the '
                    'Authorization HTTP header.'
                )
                return HttpResponse(
                    """<html><title>Auth required</title><body>
                    <h1>Authorization Required</h1><p>The archiver key must be
                     sent over the Authorization HTTP header.
                    </p></body></html>""",
                    content_type="text/html", status=401)

        auth = authorization.split()
        # Use timing-attack safe comparison of secret key
        if (len(auth) != 2 or auth[0] != 'Token' or
                not hmac.compare_digest(
                    auth[1], settings.MAILMAN_ARCHIVER_KEY)):
            return HttpResponse(
                """<html><title>Auth required</title><body>
                <h1>Authorization Required</h1><p>Please check whether
                 the MAILMAN_ARCHIVER_KEY is provided by you and it is correct.
                </p></body></html>""",
                content_type="text/html", status=401)
        return func(request, *args, **kwargs)
    return _decorator


def _get_url(mlist_fqdn, msg_id=None):
    # We can't use HttpRequest.build_absolute_uri() because the mailman API may
    # be accessed via localhost.
    # https://docs.djangoproject.com/en/dev/ref/request-response/#django.http.HttpRequest.build_absolute_uri
    # https://docs.djangoproject.com/en/dev/ref/contrib/sites/#getting-the-current-domain-for-full-urls
    # result = urljoin(public_url, urlunquote(
    #                  reverse('hk_list_overview', args=[mlist_fqdn])))
    # We use the MailDomain association from django_mailman3 to find out the
    # proper domain.
    if msg_id is None:
        url = reverse('hk_list_overview', args=[mlist_fqdn])
    else:
        msg_hash = get_message_id_hash(msg_id.strip().strip("<>"))
        url = reverse('hk_message_index', kwargs={
            "mlist_fqdn": mlist_fqdn, "message_id_hash": msg_hash})
    relative_url = unquote(url)
    mail_domain = mlist_fqdn.split("@")[1]
    try:
        domain = MailDomain.objects.get(
            mail_domain=mail_domain).site.domain
    except MailDomain.DoesNotExist:
        domain = mail_domain
    return urljoin("https://%s" % domain, relative_url)


@key_and_ip_auth
def urls(request):
    result = _get_url(request.GET["mlist"], request.GET.get("msgid"))
    return HttpResponse(json.dumps({"url": result}),
                        content_type='application/javascript')


@require_POST
@key_and_ip_auth
@csrf_exempt
def archive(request):
    mlist_fqdn = request.POST["mlist"]
    if "message" not in request.FILES:
        raise SuspiciousOperation
    msg = message_from_binary_file(
        request.FILES['message'], _class=EmailMessage, policy=default)
    try:
        add_to_list(mlist_fqdn, msg)
    except DuplicateMessage as e:
        logger.info("Duplicate email with message-id '%s'", e.args[0])
    except ValueError as e:
        logger.warning("Could not archive the email with message-id '%s': %s",
                       msg.get("Message-Id", None), e)
        return HttpResponse(json.dumps({"error": str(e)}), status=400,
                            content_type='application/javascript')
    url = _get_url(mlist_fqdn, msg['Message-Id'])
    logger.info("Archived message %s to %s", msg['Message-Id'], url)
    return HttpResponse(json.dumps({"url": url}),
                        content_type='application/javascript')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

# flake8:noqa

from .category import ThreadCategory
from .email import Attachment, Email
from .favorite import Favorite
from .mailinglist import ArchivePolicy, MailingList
from .profile import Profile
from .sender import Sender
from .tag import Tag, Tagging
from .thread import LastView, Thread
from .vote import Vote



# -*- coding: utf-8 -*-
# flake8: noqa

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0005_MailingList_list_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='thread',
            name='category',
            field=models.ForeignKey(related_name='threads', on_delete=django.db.models.deletion.SET_NULL, to='hyperkitty.ThreadCategory', null=True),
        ),
        migrations.AlterField(
            model_name='thread',
            name='starting_email',
            field=models.OneToOneField(related_name='started_thread', null=True, on_delete=django.db.models.deletion.SET_NULL, to='hyperkitty.Email'),
        ),
    ]



# -*- coding: utf-8 -*-
# flake8: noqa

from django.db import migrations, models


def remove_duplicate_persona_accounts(apps, schema_editor):
    # Some people logged in using both Persona and the other login options,
    # resulting in duplicate User accounts (with the same email address). In
    # that case, remove the Persona user and move its data over to the other
    # user instance.
    User = apps.get_model("auth", "User")
    SocialAccount = apps.get_model("socialaccount", "SocialAccount")
    Tagging = apps.get_model("hyperkitty", "Tagging")
    Favorite = apps.get_model("hyperkitty", "Favorite")
    LastView = apps.get_model("hyperkitty", "LastView")
    Vote = apps.get_model("hyperkitty", "Vote")
    query = User.objects.exclude(
            id__in=SocialAccount.objects.all().values("user_id"))
    for user in query:
        other_account = User.objects.filter(
                email=user.email
            ).exclude(
                id=user.id
            ).order_by("-last_login").first()
        if other_account is None:
            continue  # Only keep duplicate accounts.
        for tag in Tagging.objects.filter(user=user):
            tag.user = other_account
            tag.save()
        for fav in Favorite.objects.filter(user=user):
            if Favorite.objects.filter(
                    user=other_account, thread=fav.thread).exists():
                fav.delete()  # the other account has priority
            else:
                fav.user = other_account
                fav.save()
        for lv in LastView.objects.filter(user=user):
            if LastView.objects.filter(
                    user=other_account, thread=lv.thread).exists():
                lv.delete()  # the other account has priority
            else:
                lv.user = other_account
                lv.save()
        for vote in Vote.objects.filter(user=user):
            if Vote.objects.filter(
                    user=other_account, email=vote.email).exists():
                vote.delete()  # the other account has priority
            else:
                vote.user = other_account
                vote.save()
        user.delete()




class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0008_django_mailman3_profile'),
    ]

    operations = [
        migrations.RunPython(remove_duplicate_persona_accounts),
    ]



# -*- coding: utf-8 -*-
# flake8: noqa
import contextlib

from django.apps.registry import Apps
from django.db import migrations, models


class MailingListPrimaryKey(migrations.AlterField):

    def __init__(self):
        super(MailingListPrimaryKey, self).__init__(
            "mailinglist", "name", field=models.CharField(
                unique=True, max_length=254, serialize=True,
                auto_created=False)
            )

    def state_forwards(self, app_label, state):
        # django < 3.1
        if type(state.models[app_label, self.model_name_lower].fields) is list:
            state.models[app_label, self.model_name_lower].fields.insert(
                0,
                (
                    "id",
                    models.AutoField(
                        name="id",
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID'
                    )
                )
            )
        # django >= 3.1
        else:
            state.models[app_label, self.model_name_lower].fields.update(
                {
                    "id":
                    models.AutoField(
                        name="id",
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    )
                }
            )
        super(MailingListPrimaryKey, self).state_forwards(app_label, state)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        # See django.db.backends.sqlite3.schema:DatabaseSchemaEditor._remake_table()
        old_model = from_state.apps.get_model("hyperkitty", "MailingList")
        new_model = to_state.apps.get_model("hyperkitty", "MailingList")
        old_fields = [schema_editor.quote_name(f.column)
                      for f in old_model._meta.local_fields]
        body = {f.name: f for f in new_model._meta.local_fields}

        # Construct a new model for the new state
        meta_contents = {
            'app_label': old_model._meta.app_label,
            'db_table': old_model._meta.db_table,
            'apps': Apps(),
        }
        meta = type("Meta", tuple(), meta_contents)
        body['Meta'] = meta
        body['__module__'] = old_model.__module__

        temp_model = type(old_model._meta.object_name, old_model.__bases__, body)
        # We need to modify model._meta.db_table, but everything explodes
        # if the change isn't reversed before the end of this method. This
        # context manager helps us avoid that situation.
        @contextlib.contextmanager
        def altered_table_name(model, temporary_table_name):
            original_table_name = model._meta.db_table
            model._meta.db_table = temporary_table_name
            yield
            model._meta.db_table = original_table_name

        with altered_table_name(old_model, old_model._meta.db_table + "__old"):
            schema_editor.alter_db_table(
                old_model, temp_model._meta.db_table, old_model._meta.db_table)

            # Create a new table with the updated schema. We remove things
            # from the deferred SQL that match our table name, too
            schema_editor.deferred_sql = [
                x for x in schema_editor.deferred_sql
                if temp_model._meta.db_table not in x]
            schema_editor.create_model(temp_model)

            # Copy data from the old table into the new table
            schema_editor.execute(
                "INSERT INTO %s (%s) SELECT %s FROM %s ORDER BY name" % (
                schema_editor.quote_name(temp_model._meta.db_table),
                ', '.join(old_fields),
                ', '.join(old_fields),
                schema_editor.quote_name(old_model._meta.db_table),
            ))

            # Delete the old table
            schema_editor.delete_model(old_model)
        # Run deferred SQL on correct table
        for sql in schema_editor.deferred_sql:
            schema_editor.execute(sql)
        schema_editor.deferred_sql = []

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        from_model = from_state.apps.get_model(app_label, self.model_name)
        schema_editor.remove_field(from_model, from_model._meta.get_field("id"))
        super(MailingListPrimaryKey, self).database_forwards(
            app_label, schema_editor, from_state, to_state)


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0012_thread_order_null'),
    ]

    operations = [
        # Drop the constraints on thread.mailinglist_id and email.mailinglist_id
        migrations.AlterField(
            model_name='thread',
            name='mailinglist',
            field=models.CharField(db_column="mailinglist_id", max_length=254),
        ),
        migrations.AlterField(
            model_name='email',
            name='mailinglist',
            field=models.CharField(db_column="mailinglist_id", max_length=254),
        ),
        # Rebuild the mailinglist table with the new primary key.
        MailingListPrimaryKey(),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.

"""This module contains Django-haystack backend for SQLlite3."""

from haystack.backends import (
    BaseEngine, BaseSearchBackend, BaseSearchQuery, log_query)


class SqliteSearchBackend(BaseSearchBackend):

    def update(self, index, iterable, commit=True):
        return super().update(index, iterable, commit)

    def remove(self, obj_or_string):
        return super().remove(obj_or_string)

    def clear(self, models=None, commit=True):
        return super().clear(models, commit)

    @log_query
    def search(self, query_string, **kwargs):
        return super().search(query_string, **kwargs)

    def prep_value(self, value):
        return super().prep_value(value)

    def more_like_this(
            self,
            model_instance,
            additional_query_string=None,
            result_class=None
    ):
        return super().more_like_this(
            model_instance, additional_query_string, result_class)


class SqliteSearchQuery(BaseSearchQuery):

    def build_query(self):
        return super().build_query()


class SqliteEngine(BaseEngine):

    backend = SqliteSearchBackend
    query = SqliteSearchQuery



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aamir Khan <syst3m.w0rm@gmail.com>
# Author: Aurélien Bompard <abompard@fedoraproject.org>
#

from django import forms
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _


class TextInputWithButton(forms.TextInput):
    """
    Render a text field and a button following the Twitter Bootstrap
    directives: http://getbootstrap.com/components/#input-groups-buttons

    Use the 'button_text' class attribute to set the button's text.
    """

    def render(self, name, value, attrs=None, renderer=None):
        button_text = self.attrs.pop("button_text", "")
        initial_rendering = forms.TextInput.render(
                self, name, value, attrs, renderer=renderer)
        button = mark_safe(
            '<span class="input-group-btn"><button type="submit" '
            'class="btn btn-outline-primary">%s</button></span>'
            % button_text)
        return "".join([
            '<span class="input-append"><div class="input-group">',
            initial_rendering, button, '</div></span>'])


class AddTagForm(forms.Form):
    tag = forms.CharField(
        label='', help_text=None,
        widget=TextInputWithButton(
            attrs={'placeholder': _('Add a tag...'),
                   'class': 'input-medium form-control',
                   'button_text': _('Add'),
                   'title': _('use commas to add multiple tags'),
                   }
            )
        )
    action = forms.CharField(widget=forms.HiddenInput, initial="add")


class AttachmentFileInput(forms.FileInput):
    attach_first_text = _('Attach a file')
    attach_another_text = _('Attach another file')
    rm_text = _('Remove this file')
    template = """
<span class="attach-files-template">
    %(input)s <a href="#" title="%(rm_text)s">(-)</a>
</span>
<span class="attach-files"></span>
<a href="#" class="attach-files-first">%(attach_first_text)s</a>
<a href="#" class="attach-files-add">%(attach_another_text)s</a>
"""

    def render(self, name, value, attrs=None, renderer=None):
        substitutions = {
            'attach_first_text': self.attach_first_text,
            'attach_another_text': self.attach_another_text,
            'rm_text': self.rm_text,
        }
        substitutions['input'] = super(AttachmentFileInput, self).render(
            name, value, attrs, renderer=renderer)
        return mark_safe(self.template % substitutions)


class ReplyForm(forms.Form):
    newthread = forms.BooleanField(label="", required=False)
    subject = forms.CharField(
        label="", required=False,
        widget=forms.TextInput(attrs={
            'placeholder': 'New subject', 'class': 'form-control'}))
    message = forms.CharField(
        label="",
        widget=forms.Textarea(attrs={'class': 'form-control'}))
    sender = forms.ChoiceField(
        label="", required=False,
        widget=forms.Select(attrs={'class': 'form-select input-sm'}))
    #  attachment = forms.FileField(required=False, widget=AttachmentFileInput)


class PostForm(forms.Form):

    subject = forms.CharField()
    message = forms.CharField(widget=forms.Textarea)
    sender = forms.ChoiceField(
        label="", required=False,
        widget=forms.Select(attrs={'class': 'form-select input-sm'}))
    # attachment = forms.FileField(required=False, label="",
    #                              widget=AttachmentFileInput)


class CategoryForm(forms.Form):
    category = forms.ChoiceField(label="", required=False)


class MessageDeleteForm(forms.Form):
    email = forms.ModelMultipleChoiceField(
        queryset=None, widget=forms.ModelMultipleChoiceField.hidden_widget,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aamir Khan <syst3m.w0rm@gmail.com>
#


class Email(object):
    """ Email class containing the information needed to store and
    display email threads.
    """

    def __init__(self):
        """ Constructor.
        Instanciate the default attributes of the object.
        """
        self.email_id = ''
        self.title = ''
        self.body = ''
        self.tags = []
        self.category = 'question'
        self.category_tag = None
        self.participants = set(['Pierre-Yves Chibon'])
        self.answers = []
        self.liked = 0
        self.author = ''
        self.avatar = None
        self.age = '6 days'


class Author(object):
    """ Author class containing the information needed to get the top
    author of the month!
    """

    def __init__(self):
        """ Constructor.
        Instanciate the default attributes of the object.
        """
        self.name = None
        self.kudos = 0
        self.avatar = None


def get_email_tag(tag):
    threads = generate_random_thread()
    output = []
    for email in threads:
        if tag in email.tags or tag in email.category:
            output.append(email)
        elif email.category_tag and tag in email.category_tag:
            output.append(email)
    return output


def generate_thread_per_category():
    threads = generate_random_thread()
    categories = {}
    for thread in threads:
        category = thread.category
        if thread.category_tag:
            category = thread.category_tag
        if category in categories.keys():
            categories[category].append(thread)
        else:
            categories[category] = [thread]
    return categories


def generate_top_author():
    authors = []

    author = Author()
    author.name = 'Pierre-Yves Chibon'
    author.avatar = ('https://secure.gravatar.com/avatar/'
                     '072b4416fbfad867a44bc7a5be5eddb9')
    author.kudos = 3
    authors.append(author)

    author = Author()
    author.name = 'Stanislav Ochotnický'
    author.avatar = 'http://sochotni.fedorapeople.org/sochotni.jpg'
    author.kudos = 4
    authors.append(author)

    author = Author()
    author.name = 'Toshio Kuratomi'
    author.avatar = ('https://secure.gravatar.com/avatar/'
                     '7a9c1d88f484c9806bceca0d6d91e948')
    author.kudos = 5
    authors.append(author)

    return authors


def generate_random_thread():
    threads = []

    # 1
    email = Email()
    email.email_id = 1
    email.title = 'Headsup! krb5 ccache defaults are changing in Rawhide'
    email.age = '6 days'
    email.body = '''Dear fellow developers,
with the upcoming Fedora 18 release (currently Rawhide) we are going to
change the place where krb5 credential cache files are saved by default.

The new default for credential caches will be the /run/user/username directory.
'''
    email.tags.extend(['rawhide', 'krb5'])
    email.participants = set([
        'Stephen Gallagher', 'Toshio Kuratomi', 'Kevin Fenzi', 'Seth Vidal',
        ])
    email.answers.extend([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    email.liked = 1
    email.author = 'Stephen Gallagher'
    email.avatar = 'http://fedorapeople.org/~sgallagh/karrde712.png'
    threads.append(email)

    # 2
    email = Email()
    email.email_id = 2
    email.title = 'Problem in packaging kicad'
    email.age = '6 days'
    email.body = '''Paragraph 1: Lorem ipsum dolor sit amet, consectetur
adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore
magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco
laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in
reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla
pariatur. '''
    email.tags.extend(['packaging', 'kicad'])
    email.participants = set([
        'Pierre-Yves Chibon', 'Tom "spot" Callaway', 'Toshio Kuratomi',
        'Kevin Fenzi'])
    email.answers.extend([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    email.liked = 0
    email.author = 'Pierre-Yves Chibon'
    email.avatar = ('https://secure.gravatar.com/avatar/'
                    '072b4416fbfad867a44bc7a5be5eddb9')
    threads.append(email)

    # 3
    email = Email()
    email.email_id = 3
    email.title = 'Update Java Guideline'
    email.age = '6 days'
    email.body = '''Paragraph 1: Lorem ipsum dolor sit amet, consectetur
adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore
magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco
laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in
reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla
pariatur. '''
    email.tags.extend(['rawhide', 'krb5'])
    email.participants = set([
        'Stanislav Ochotnický', 'Tom "spot" Callaway', 'Stephen Gallagher',
        'Jason Tibbitts', 'Rex Dieter', 'Toshio Kuratomi'])
    email.answers.extend([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
                          16, 17, 18, 19])
    email.liked = 5
    email.category = 'todo'
    email.author = 'Stanislav Ochotnický'
    email.avatar = 'http://sochotni.fedorapeople.org/sochotni.jpg'
    threads.append(email)

    # 4
    email = Email()
    email.email_id = 4
    email.title = 'Agenda for the next Board Meeting'
    email.age = '6 days'
    email.body = '''Paragraph 1: Lorem ipsum dolor sit amet, consectetur
adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore
magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco
laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in
reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla
pariatur. '''
    email.tags.extend(['agenda', 'board'])
    email.participants = set([
        'Toshio Kuratomi', 'Tom "spot" Callaway', 'Robyn Bergeron',
        'Max Spevack'])
    email.answers.extend([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    email.liked = 20
    email.category = 'agenda'
    email.author = 'Toshio Kuratomi'
    email.avatar = ('https://secure.gravatar.com/avatar/'
                    '7a9c1d88f484c9806bceca0d6d91e948')
    threads.append(email)

    # 5
    email = Email()
    email.email_id = 5
    email.title = 'I told you so! '
    email.age = '6 days'
    email.body = '''Paragraph 1: Lorem ipsum dolor sit amet, consectetur
adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore
magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco
laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in
reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla
pariatur. '''
    email.tags.extend(['systemd', 'mp3', 'pulseaudio'])
    email.participants = set(['Pierre-Yves Chibon'])
    email.answers.extend([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    email.liked = 0
    email.author = 'Pierre-Yves Chibon'
    email.avatar = ('https://secure.gravatar.com/avatar/'
                    '072b4416fbfad867a44bc7a5be5eddb9')
    email.category = 'shut down'
    email.category_tag = 'dead'
    threads.append(email)

    return threads



# -*- coding: utf-8 -*-
# flake8: noqa

from django.db import migrations, models


def populate_list_id(apps, schema_editor):
    MailingList = apps.get_model("hyperkitty", "MailingList")
    for ml in MailingList.objects.filter(list_id=None):
        ml.list_id = ml.name.replace("@", ".")
        ml.save()


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0004_archived_date_and_subject'),
    ]

    operations = [
        migrations.AddField(
            model_name='mailinglist',
            name='list_id',
            field=models.CharField(max_length=254, null=True, unique=True),
        ),
        migrations.RunPython(populate_list_id),
    ]



# -*- coding: utf-8 -*-
# flake8: noqa

from django.db import migrations, models


def find_starting_email(apps, schema_editor):
    # We can't import the Thread model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    Thread = apps.get_model("hyperkitty", "Thread")
    Email = apps.get_model("hyperkitty", "Email")
    for thread in Thread.objects.all():
        try:
            thread.starting_email = thread.emails.get(parent_id__isnull=True)
        except Email.DoesNotExist:
            thread.starting_email = thread.emails.order_by("date").first()
        thread.save()

class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0002_auto_20150311_0913'),
    ]

    operations = [
        migrations.AddField(
            model_name='thread',
            name='starting_email',
            field=models.OneToOneField(related_name='started_thread',
                                       null=True, to='hyperkitty.Email',
                                       on_delete=models.CASCADE),
        ),
        migrations.RunPython(find_starting_email),
    ]



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from rest_framework import serializers

from hyperkitty.models import Sender


class OptionalHyperlinkedRelatedField(serializers.HyperlinkedRelatedField):
    def to_representation(self, obj):
        value = getattr(obj, self.lookup_field)
        if value is None:
            return None
        return super(
            OptionalHyperlinkedRelatedField, self
            ).to_representation(obj)


class SenderSerializer(serializers.HyperlinkedModelSerializer):
    emails = OptionalHyperlinkedRelatedField(
        view_name='hk_api_sender_email_list', read_only=True,
        lookup_field="mailman_id", source="*", required=False)
    # emails is None if mailmain_id is None
    address = serializers.SerializerMethodField()

    class Meta:
        model = Sender
        fields = ("address", "mailman_id", "address", "emails")
        lookup_field = "address"

    def get_address(self, obj):
        return obj.address.replace("@", " (a) ")



# -*- coding: utf-8 -*-
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
# Modified by Mark Sapiro <mark@msapiro.net>
#
from django.core.management.base import CommandError
from django.http.response import Http404
from django.shortcuts import get_object_or_404

from haystack import indexes
from haystack.management.commands.update_index import (
    Command as UpdateIndexCommand)
from haystack.query import SearchQuerySet

from hyperkitty.models import Email, MailingList


# Create a global for the listname.
LISTNAME = None


class EmailIndex(indexes.SearchIndex, indexes.Indexable):

    text = indexes.CharField(document=True, use_template=True)
    mailinglist = indexes.CharField(model_attr='mailinglist__name')
    subject = indexes.CharField(model_attr='subject', boost=1.25,
                                use_template=True)
    date = indexes.DateTimeField(model_attr='date')
    sender = indexes.CharField(
        model_attr='sender_name', null=True, boost=1.125)
    tags = indexes.MultiValueField(
        model_attr='thread__tags__name', null=True, boost=1.25)
    archived_date = indexes.DateTimeField(model_attr='archived_date')

    def get_model(self):
        return Email

    def get_updated_field(self):
        return 'archived_date'

    def index_queryset(self, using=None):
        if LISTNAME is None:
            return self.get_model().objects.all()
        else:
            return self.get_model().objects.filter(
                mailinglist__name=LISTNAME)

    def load_all_queryset(self):
        # Pull other objects related to the Email in search results.
        return self.get_model().objects.all().select_related(
            "sender", "thread")


def update_index(remove=False, listname=None, verbosity=0):
    """
    Update the search index with the new emails since the last index update
    or if listname is provided, with all emails from that list.

    Setting remove to True is extremely slow, it needs to scan the entire
    index and database. It takes about 15 minutes on Fedora's lists, so it is
    not fit for a frequent operation.

    The listname option is intended to update a single list after importing
    that list's archives.  Doing the entire archive takes way too long and
    doing a 'since' doesn't get the old imported posts.
    """
    global LISTNAME
    LISTNAME = listname
    update_cmd = UpdateIndexCommand()
    if LISTNAME is None:
        # Find the last email in the index:
        try:
            last_email = SearchQuerySet().latest('archived_date')
        except Exception:
            # Different backends can raise different exceptions unfortunately
            update_cmd.start_date = None
        else:
            update_cmd.start_date = last_email.archived_date
    else:
        # Is this a valid list?
        try:
            get_object_or_404(MailingList, name=listname)
        except Http404 as e:
            raise CommandError('{}: {}'.format(listname, e))
        # set the start date to None to do the whole list.
        update_cmd.start_date = None
    # set defaults
    update_cmd.verbosity = verbosity
    update_cmd.batchsize = None
    update_cmd.end_date = None
    update_cmd.workers = 0
    update_cmd.commit = True
    update_cmd.remove = remove
    try:
        from haystack.management.commands.update_index import (
            DEFAULT_MAX_RETRIES)
    except ImportError:
        pass
    else:
        update_cmd.max_retries = DEFAULT_MAX_RETRIES
    update_cmd.update_backend("hyperkitty", "default")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>

"""
Import new lists from Mailman over REST.
Those lists have no archived email yet.
"""

from django_extensions.management.jobs import BaseJob

from hyperkitty.lib.mailman import get_new_lists_from_mailman


class Job(BaseJob):
    help = "Import new lists from Mailman"
    when = "hourly"

    def execute(self):
        get_new_lists_from_mailman()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from django.conf import settings
from django.db import models


class Vote(models.Model):
    """
    A User's vote on a message
    """
    email = models.ForeignKey(
        "Email", related_name="votes", on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL,
                             related_name="votes", on_delete=models.CASCADE)
    value = models.SmallIntegerField(db_index=True)

    class Meta:
        unique_together = ("email", "user")

    def on_post_save(self):
        self.email.on_vote_added(self)
        self.email.thread.on_vote_added(self)
        self.email.mailinglist.on_vote_added(self)

    def on_post_delete(self):
        self.email.on_vote_deleted(self)
        self.email.thread.on_vote_deleted(self)
        self.email.mailinglist.on_vote_deleted(self)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aamir Khan <syst3m.w0rm@gmail.com>
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from django.conf.urls import include
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import path, re_path
from django.views.generic.base import TemplateView

from hyperkitty.api import email as api_email
from hyperkitty.api import mailinglist as api_mailinglist
from hyperkitty.api import tag as api_tag
from hyperkitty.api import thread as api_thread
from hyperkitty.feed import MailingListFeed
from hyperkitty.lib.view_helpers import check_mlist_private
from hyperkitty.views import (
    accounts, categories, compat, index, mailman, message, mlist, search,
    tags, thread, users)


# flake8: noqa


# List archives and overview
list_patterns = [
    re_path(r'^(?P<year>\d{4})/(?P<month>\d\d?)/(?P<day>\d\d?)/$',
        mlist.archives, name='hk_archives_with_day'),
    re_path(r'^(?P<year>\d{4})/(?P<month>\d\d?)/$',
        mlist.archives, name='hk_archives_with_month'),
    path('latest', mlist.archives, name='hk_archives_latest'),
    path('', mlist.overview, name='hk_list_overview'),
    path('recent-activity', mlist.recent_activity, name='hk_list_recent_activity'),
    path('recent-threads', mlist.overview_recent_threads, name='hk_list_overview_recent_threads'),
    path('pop-threads', mlist.overview_pop_threads, name='hk_list_overview_pop_threads'),
    path('top-threads', mlist.overview_top_threads, name='hk_list_overview_top_threads'),
    path('favorites', mlist.overview_favorites, name='hk_list_overview_favorites'),
    path('posted-to', mlist.overview_posted_to, name='hk_list_overview_posted_to'),
    path('top-posters', mlist.overview_top_posters, name='hk_list_overview_top_posters'),
    re_path(r'^export/(?P<filename>[^/]+)\.mbox.gz$',
        mlist.export_mbox, name='hk_list_export_mbox'),
    path('delete/', mlist.delete, name='hk_list_delete'),
    path('feed/', check_mlist_private(MailingListFeed()), name='hk_list_feed'),
]


# Messages
message_patterns = [
    path('', message.index, name='hk_message_index'),
    path('attachment/<int:counter>/<path:filename>', message.attachment, name='hk_message_attachment'),
    path('vote', message.vote, name='hk_message_vote'),
    path('reply', message.reply, name='hk_message_reply'),
    path('delete', message.delete, name='hk_message_delete'),
]


# Threads
thread_patterns = [
    path('', thread.thread_index, name='hk_thread'),
    path('replies', thread.replies, name='hk_thread_replies'),
    path('tags', thread.tags, name='hk_tags'),
    path('suggest-tags', thread.suggest_tags, name='hk_suggest_tags'),
    path('favorite', thread.favorite, name='hk_favorite'),
    path('category', thread.set_category, name='hk_thread_set_category'),
    path('reattach', thread.reattach, name='hk_thread_reattach'),
    path('reattach-suggest', thread.reattach_suggest, name='hk_thread_reattach_suggest'),
    path('delete', message.delete, name='hk_thread_delete'),
]


# REST API
api_list_patterns = [
    path('', api_mailinglist.MailingListDetail.as_view(), name="hk_api_mailinglist_detail"),
    path('threads/', api_thread.ThreadList.as_view(), name="hk_api_thread_list"),
    path('thread/<str:thread_id>/', api_thread.ThreadDetail.as_view(), name="hk_api_thread_detail"),
    path('emails/', api_email.EmailList.as_view(), name="hk_api_email_list"),
    re_path(r'^email/(?P<message_id_hash>.*)/$',
        api_email.EmailDetail.as_view(), name="hk_api_email_detail"),
    path('thread/<str:thread_id>/emails/', api_email.EmailList.as_view(), name="hk_api_thread_email_list"),
]
api_patterns = [
    path('', TemplateView.as_view(template_name="hyperkitty/api.html")),
    path('lists/', api_mailinglist.MailingListList.as_view(), name="hk_api_mailinglist_list"),
    re_path(r'^list/(?P<mlist_fqdn>[^/@]+@[^/@]+)/', include(api_list_patterns)),
    path('sender/<str:mailman_id>/emails/', api_email.EmailListBySender.as_view(), name="hk_api_sender_email_list"),
    path('tags/', api_tag.TagList.as_view(), name="hk_api_tag_list"),
]


urlpatterns = [
    # Index
    path('', index.index, name='hk_root'),
    path('find-list', index.find_list, name='hk_find_list'),

    # User profile
    path('profile/', include([
        path('', accounts.user_profile, name='hk_user_profile'),
        path('favorites', accounts.favorites, name='hk_user_favorites'),
        path('last_views', accounts.last_views, name='hk_user_last_views'),
        path('votes', accounts.votes, name='hk_user_votes'),
        path('subscriptions', accounts.subscriptions,
            name='hk_user_subscriptions'),
    ])),

    # Users
    path('users/', users.users, name='hk_users_overview'),
    path('users/<str:user_id>/', accounts.public_profile, name='hk_public_user_profile'),
    path('users/<str:user_id>/posts', accounts.posts, name='hk_user_posts'),

    # List archives and overview
    re_path(r'^list/(?P<mlist_fqdn>[^/@]+@[^/@]+)/', include(list_patterns)),

    # Messages
    re_path(r'^list/(?P<mlist_fqdn>[^/@]+@[^/@]+)/message/'
        r'(?P<message_id_hash>\w+)/', include(message_patterns)),
    re_path(r'^list/(?P<mlist_fqdn>[^/@]+@[^/@]+)/message/new$',
        message.new_message, name='hk_message_new'),

    # Threads
    re_path(r'^list/(?P<mlist_fqdn>[^/@]+@[^/@]+)/thread/(?P<threadid>\w+)/',
        include(thread_patterns)),

    # Search
    path('search', search.search, name='hk_search'),

    # Categories and Tags
    path('categories/', categories.categories, name='hk_categories_overview'),
    path('tags/', tags.tags, name='hk_tags_overview'),

    # Mailman archiver API
    path('api/mailman/urls', mailman.urls, name='hk_mailman_urls'),
    path('api/mailman/archive', mailman.archive, name='hk_mailman_archive'),

    # REST API
    path('api/', include(api_patterns)),

    # Mailman 2.X compatibility
    re_path(r'^listinfo/?$', compat.summary),
    re_path(r'^listinfo/(?P<list_name>[^/]+)/?$', compat.summary),
    re_path(r'^pipermail/(?P<list_name>[^/]+)/?$', compat.summary),
    re_path(r'^pipermail/(?P<list_name>[^/]+)/(?P<year>\d\d\d\d)-(?P<month_name>\w+)/?$', compat.arch_month),
    re_path(r'^pipermail/(?P<list_name>[^/]+)/(?P<year>\d\d\d\d)-(?P<month_name>\w+)/(?P<summary_type>[a-z]+)\.html$', compat.arch_month),
    re_path(r'^pipermail/(?P<list_name>[^/]+)/(?P<year>\d\d\d\d)-(?P<month_name>\w+)\.txt.gz', compat.arch_month_mbox),
    #url(r'^pipermail/(?P<list_name>[^/]+)/(?P<year>\d\d\d\d)-(?P<month_name>\w+)/(?P<msg_num>\d+)\.html$', compat.message),
    re_path(r'^list/(?P<list_name>[^@]+)@[^/]+/(?P<year>\d\d\d\d)-(?P<month_name>\w+)/?$', compat.arch_month),
    #url(r'^list/(?P<list_name>[^@]+)@[^/]+/(?P<year>\d\d\d\d)-(?P<month_name>\w+)/(?P<msg_num>\d+)\.html$', compat.message),

    # URL compatibility with previous versions
    re_path(r'^list/(?P<list_id>[^@/]+)/', compat.redirect_list_id),
    path('lists/', compat.redirect_lists),

]
#) + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)
urlpatterns += staticfiles_urlpatterns()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import logging
from urllib.error import HTTPError

from django.core.cache import cache

from django_mailman3.lib.mailman import get_mailman_client
from mailmanclient import MailmanConnectionError


logger = logging.getLogger(__name__)


class ModeratedListException(Exception):
    pass


class AddressBannedFromList(Exception):
    pass


def subscribe(list_id, user, email=None, display_name=None):
    """Subscribe a user to a given MailingList.

    :param list_id: List to subscribe the user to.
    :param user: Django user to subscribe.
    :param email: Email address of the Django user. If not provided
        the default email of the Django user is used.
    :param display_name: The Display name of the user being subscribed,
        if not provided, a join of First and Last name of the Django
        user is used.

    :returns: Boolean indicating if the user was successfully subscribed
        to the list or not.
    """
    if email is None:
        email = user.email
    if display_name is None:
        display_name = "%s %s" % (user.first_name, user.last_name)
    client = get_mailman_client()
    rest_list = client.get_list(list_id)
    subscription_policy = rest_list.settings.get(
        "subscription_policy", "moderate")
    # Add a flag to return that would tell the user they have been subscribed
    # to the current list.
    subscribed_now = False
    try:
        member = rest_list.get_member(email)
    except ValueError:
        # We don't want to bypass moderation, don't subscribe. Instead
        # raise an error so that it can be caught to show the user
        if subscription_policy in ("moderate", "confirm_then_moderate"):
            raise ModeratedListException(
                "This list is moderated, please subscribe"
                " to it before posting.")

        # not subscribed yet, subscribe the user without email delivery
        try:
            member = rest_list.subscribe(
                email, display_name, pre_verified=True, pre_confirmed=True)
        except HTTPError as e:
            if e.code == 409:
                logger.info("Subscription for %s to %s is already pending",
                            email, list_id)
                return subscribed_now
            elif e.code == 400 and e.reason == 'Membership is banned':
                # Using the message string to actually do something here is
                # not the best idea, but we don't have any better way to
                # handle errors right now.
                # Ideally, API will assign different error codes to different
                # types of errors and we can use an internal no. to signal
                # various different errors.
                logger.info("%s is banned from %s", email, list_id)
                raise AddressBannedFromList(
                    "Address {} is banned from {}".format(email, list_id))
            else:
                raise
        # The result can be a Member object or a dict if the subscription can't
        # be done directly, or if it's pending, or something else.
        # Broken API :-(
        if isinstance(member, dict):
            logger.info("Subscription for %s to %s is pending",
                        email, list_id)
            return subscribed_now
        member.preferences["delivery_status"] = "by_user"
        member.preferences.save()
        subscribed_now = True
        cache.delete("User:%s:subscriptions" % user.id, version=2)
        logger.info("Subscribing %s to %s on first post",
                    email, list_id)

    return subscribed_now


def get_new_lists_from_mailman():
    """Poll Mailman's API and get all the new MailingLists.

    This is needed so that Hyperkitty can synchronize the internal
    state with Mailman Core. Ideally, HK is getting right signals from
    Postorius or email archiving is causing the new lists to be imported
    from the API on demand.

    This is typically invoked from one of the scheduled cron jobs.
    """
    from hyperkitty.models import MailingList
    mmclient = get_mailman_client()
    page_num = 0
    while page_num < 10000:  # Just for security
        page_num += 1
        try:
            mlist_page = mmclient.get_list_page(count=10, page=page_num)
        except MailmanConnectionError:
            break
        except HTTPError:
            break  # can't update at this time
        for mm_list in mlist_page:
            if MailingList.objects.filter(name=mm_list.fqdn_listname).exists():
                continue
            if mm_list.settings["archive_policy"] == "never":
                continue  # Should we display those lists anyway?
            logger.info("Imported the new list %s from Mailman",
                        mm_list.fqdn_listname)
            mlist = MailingList.objects.create(name=mm_list.fqdn_listname)
            mlist.update_from_mailman()
        if not mlist_page.has_next:
            break


def import_list_from_mailman(list_id):
    """Given a single MailingList, import details from Core.

    :param list_id: List id for MailingList to import.
    """
    from hyperkitty.models import MailingList
    mmclient = get_mailman_client()
    try:
        mm_list = mmclient.get_list(list_id)
    except (MailmanConnectionError, HTTPError):
        return
    mlist, created = MailingList.objects.get_or_create(
        name=mm_list.fqdn_listname)
    if created:
        logger.info("Imported the new list %s from Mailman",
                    mm_list.fqdn_listname)
    mlist.update_from_mailman()


def sync_with_mailman(overwrite=False):
    """Sync all the details from Mailman.

    This is invoked from Cron job/hyperkitty_import or hyperkitty_sync command.

    :param overwrite: Overwrite the existing Senders.
    """
    from hyperkitty.models import MailingList, Sender
    for mlist in MailingList.objects.all():
        mlist.update_from_mailman()
    # Now sync Sender.mailman_id with Mailman's User.user_id
    # There can be thousands of senders, break into smaller chunks to avoid
    # hogging up the memory
    buffer_size = 1000
    query = Sender.objects.all()
    if not overwrite:
        query = query.filter(mailman_id__isnull=True)
    prev_count = query.count()
    lower_bound = 0
    upper_bound = buffer_size
    while True:
        try:
            for sender in query.all()[lower_bound:upper_bound]:
                sender.set_mailman_id()
        except MailmanConnectionError:
            break  # Can't refresh at this time
        count = query.count()
        if count == 0:
            break  # all done
        if count == prev_count:
            # no improvement...
            if count < upper_bound:
                break  # ...and all users checked
            else:
                # there may be some more left
                lower_bound = upper_bound
                upper_bound += buffer_size
        prev_count = count
        logger.info("%d emails left to refresh, checked %d",
                    count, lower_bound)



import re

from django.conf import settings

import mistune
from mistune.plugins.url import url
from mistune.util import safe_entity


class MyRenderer(mistune.HTMLRenderer):
    """Modified HTML renderer.

    This renderer changes a couple of things in the default renderer:
    - Add a quoted-switch to toggle blockquotes.
    - Add the marker along with the emphasis and strong.
    - Optionally render Image tags depending on RENDER_INLINE_IMAGE setting,
      which is off by default to prevent remote images being rendered inline
      as they are used for tracking and can be privacy violating.
    """

    def block_quote(self, text):
        """Returns a rendered blockquote with quote-text class and a
        quote-switched classed hyperlink that can collapse the next quote-text
        using JS.
        """
        return (
            f'<div class="quoted-switch"><a href="#">...</a></div>'
            f'<blockquote class="blockquote quoted-text">{text}</blockquote>')

    def emphasis(self, text, marker):
        """Emphasis with marker included."""
        return super().emphasis(marker + text + marker)

    def strong(self, text, marker):
        """Strong with marker included."""
        return super().strong(marker + text + marker)

    def text(self, text, marker=None):
        # We had to override this because for some reason, text inside emphasis
        # and strong will receive the attrs of parents too, like "marker" in
        # our case. We don't really need to do anything with it.
        return text

    def _md_style_img(self, src, title, alt):
        """Markdown syntax for images. """
        title = f'"{title}"' if title else ''
        return '![{alt}]({src} {title})'.format(
            src=src, title=title, alt=alt)

    def image(self, alt, url, title=None):
        """Render image if configured to do so.

        HYPERKITTY_RENDER_INLINE_IMAGE configuration allows for
        rendering of inline images in the markdown. This is disabled by
        default since embeded images can cause problems.
        """
        if getattr(settings, 'HYPERKITTY_RENDER_INLINE_IMAGE', False):
            return super().image(alt, url, title)
        return self._md_style_img(url, title, alt)

    def link(self, text, url, title=None):
        """URL link renderer that truncates the length of the URL.

        This only does it for the URLs that are not hyperlinks by just literal
        URLs (text=None) so text is same as URL.
        It also adds target=“_blank” so that the URLs open in a new tab.
        """
        # text can be none of same as url in case of autolink parsing. This
        # will truncate the length of the URL in both cases but preserve
        # the actual URL destination in the hyperlink.
        if text is None or text == url:
            text = url
            if len(text) > 76:
                text = url[:76] + '...'

        s = '<a target="_blank" href="' + self.safe_url(url) + '"'
        if title:
            s += ' title="' + safe_entity(title) + '"'
        return s + '>' + (text or url) + '</a>'


class InlineParser(mistune.inline_parser.InlineParser):
    """Modified parser that returns the marker along with emphasized text.
    We do this so we can apply styling without modifying the text itself, like
    removing `**` or `__`. Since both the markers have same styling effect,
    the renderer currently doesn’t get which marker was used, it just gets
    ‘emphasis’ or ‘strong’ node.
    We also null the ESCAPE string to prevent removing backslash escapes
    without unwanted side effects from removing the 'escapre' rule.
    """

    def parse_emphasis(self, m, state):
        end_pos = super().parse_emphasis(m, state)
        last_token = state.tokens[-1].copy()
        marker = m.group(0)
        last_token['attrs'] = {'marker': marker}
        state.tokens[-1] = last_token
        return end_pos

    def parse_escape(self, m, state):
        # Upstream version will use `unescape_char` on the text, which
        # removes the `\` in _some_ cases. This will remove the use of
        # unescape char since we don't want to do escaping at all.
        state.append_token({
            'type': 'text',
            'raw': m.group(0),
        })
        return m.end()


def remove_header_rules(rules):
    rules = list(rules)
    for rule in ('setex_header', 'axt_heading'):
        if rule in rules:
            rules.remove(rule)
    return rules


class BlockParser(mistune.block_parser.BlockParser):
    """A copy of Mistune's block parser with header parsing rules removed."""
    DEFAULT_RULES = remove_header_rules(
        mistune.block_parser.BlockParser.DEFAULT_RULES)


OUTLOOK_REPLY_PATTERN = (
    r'^-------- Original message --------\n'
    r'(?P<reply_text>[\s\S]+)'                    # everything after newline
)


def parse_outlook_reply(block, m, state):
    """Parser for outlook style replies."""
    text = m.group('reply_text')
    reply_token = '-------- Original message --------\n'
    state.append_token({
        'type': 'block_quote',
        'children': [{'type': 'paragraph', 'text': reply_token + text}],
        })
    return m.end() + 1


def plugin_outlook_reply(md):
    md.block.register(
        'outlook_reply', OUTLOOK_REPLY_PATTERN, parse_outlook_reply)


# Signature Plugin looks for signature pattern in email content and converts it
# into a muted text.
SIGNATURE_PATTERN = re.compile(
    r'^-- \n'        # --<empty space><newline>
    r'([\s\S]+)',     # Everything after newline.,
    re.M
)


def parse_signature(block, m, state):
    """Parser for signature type returns an AST node."""
    return {'type': 'signature', 'text': m.group(0), }


def render_html_signature(signature_text):
    """Render a signature as HTML."""
    return f'<div class="text-muted">{signature_text}</div>'


def plugin_signature(md):
    """Signature Plugin looks for signature pattern in email content and
    converts it into a muted text.

    It only provides an HTML renderer because that is the only one needed.
    """
    md.block.register('signature', SIGNATURE_PATTERN, parse_signature)

    # md.block.rules.insert(0,  'signature')
    if md.renderer and md.renderer.NAME == 'html':
        md.renderer.register('signature', render_html_signature)


def plugin_disable_markdown(md):
    """This plugin disables most of the rules in mistune.

    This uses mistune to do only block_quote parsing and then some inline rules
    to render an email as simple text instead of rich text. This makes such
    that the code path is same for MailingLists that do and don't enable the
    markdown rendering.
    """
    md.block.rules = ['block_quote']
    md.block.block_quote_rules = ['block_quote', 'blank_line']
    md.block.list_rules = []
    md.inline.rules = ['inline_html', 'auto_link']


# PGP Signature plugin parses inline pgp signatures from Emails and converts
# them into quoted text so that they can be collapsed.
PGP_SIGNATURE_MATCH = re.compile(
    r"^.*(BEGIN PGP SIGNATURE).*$"
    r"([\s\S]+)"
    r"^.*(END PGP SIGNATURE).*$",
    re.M)


def parse_pgp_signature(block, m, state):
    """Return a parsed pgp node."""
    return {'type': 'pgp', 'text': m.group(0)}


def render_pgp_signature(text):
    """Render pgp signature with a quote-switch and quoted-text.

    This allows collapsing pgp signatures.
    """
    return (f'<div class="quoted-switch"><a href="#">...PGP SIGNATURE...</a>'
            f'</div><div class="pgp quoted-text">{text}</div>')


def plugin_pgp_signature(md):
    """PGP signature plugin adds support to collapse pgp signature in emails

    It parses BEGIN PGP SIGNATURE and END PGP SIGNATURE and collapses content
    in between them.
    """
    md.block.register('pgp', PGP_SIGNATURE_MATCH, parse_pgp_signature)
    # md.block.rules.append('pgp')
    if md.renderer and md.renderer.NAME == 'html':
        md.renderer.register('pgp', render_pgp_signature)


renderer = MyRenderer()
markdown_renderer = mistune.Markdown(
    renderer=renderer,
    inline=InlineParser(hard_wrap=False),
    block=BlockParser(),
    plugins=[
        plugin_outlook_reply,
        plugin_pgp_signature,
        plugin_signature,
        url,
        ])


# The only difference between the markdown and this renderer is
# plugin_disable_markdown which disables all but a few markdown processing
# rules that results in a regularly formatted email.
text_renderer = mistune.Markdown(
    renderer=renderer,
    inline=InlineParser(hard_wrap=False),
    block=BlockParser(),
    plugins=[plugin_disable_markdown,
             plugin_pgp_signature,
             plugin_signature,
             url,
             ])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from django.conf import settings
from django.db import models


class Favorite(models.Model):
    thread = models.ForeignKey(
        "Thread", related_name="favorites", on_delete=models.CASCADE)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="favorites", on_delete=models.CASCADE)

    def __str__(self):
        return "%s is a favorite of %s" % (
            str(self.thread), str(self.user))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#
from django.core.cache import cache


class CachedValue(object):

    cache_key = None
    timeout = None

    def _get_cache_key(self, *args, **kwargs):
        if self.cache_key is not None:
            return self.cache_key
        raise NotImplementedError

    def get_value(self, *args, **kwargs):
        """Get the value that must be cached."""
        raise NotImplementedError

    def warm_up(self, *args, **kwargs):
        """Stores the value in the cache if it is not there already."""
        if cache.get(self._get_cache_key(*args, **kwargs)) is None:
            self.rebuild(*args, **kwargs)

    def rebuild(self, *args, **kwargs):
        """Overwrite the value in the cache."""
        value = self.get_value(*args, **kwargs)
        cache.set(self._get_cache_key(*args, **kwargs), value, self.timeout)
        return value

    def get_or_set(self, *args, **kwargs):
        """Return the cached value, rebuilding the cache if necessary."""
        value = cache.get(self._get_cache_key(*args, **kwargs))
        if value is None:
            value = self.rebuild(*args, **kwargs)
        return value

    def __call__(self, *args, **kwargs):
        return self.get_or_set(*args, **kwargs)


class ModelCachedValue(CachedValue):

    def __init__(self, instance):
        self.instance = instance

    def _get_cache_key(self, *args, **kwargs):
        if self.cache_key is not None:
            return "%s:%s:%s" % (
                self.instance.__class__.__name__,
                self.instance.pk,
                self.cache_key)
        raise NotImplementedError


class VotesCachedValue(ModelCachedValue):

    cache_key = "votes"

    def get_value(self):
        from .email import Email
        from .thread import Thread
        from .vote import Vote
        if isinstance(self.instance, Thread):
            filters = {"email__thread_id": self.instance.id}
        elif isinstance(self.instance, Email):
            filters = {"email_id": self.instance.id}
        else:
            ValueError("The 'votes' cached value only accepts 'Email' "
                       "and 'Thread' instance")
        votes = list(Vote.objects.filter(**filters).values_list(
            "value", flat=True))
        return (
                len([v for v in votes if v == 1]),
                len([v for v in votes if v == -1]),
            )

    def get_or_set(self):
        votes = super(VotesCachedValue, self).get_or_set()
        likes, dislikes = votes
        # XXX: use an Enum?
        if likes - dislikes >= 10:
            status = "likealot"
        elif likes - dislikes > 0:
            status = "like"
        else:
            status = "neutral"
        return {"likes": likes, "dislikes": dislikes, "status": status}
